import json
import os
import sys
import threading
import time

from io import BytesIO
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, TYPE_CHECKING, Type, cast

from ..debug import debug
from ..packaging_sys import ContentType, MetaflowCodeContent
from ..packaging_sys.backend import PackagingBackend
from ..packaging_sys.tar_backend import TarPackagingBackend
from ..packaging_sys.v1 import MetaflowCodeContentV1
from ..packaging_sys.utils import suffix_filter, walk
from ..metaflow_config import DEFAULT_PACKAGE_SUFFIXES
from ..exception import MetaflowException
from ..user_configs.config_parameters import dump_config_values
from .. import R

DEFAULT_SUFFIXES_LIST = DEFAULT_PACKAGE_SUFFIXES.split(",")


if TYPE_CHECKING:
    import metaflow.datastore


class NonUniqueFileNameToFilePathMappingException(MetaflowException):
    headline = "Non-unique file path for a file name included in code package"

    def __init__(self, filename, file_paths, lineno=None):
        msg = (
            "Filename %s included in the code package includes multiple different "
            "paths for the same name : %s.\n"
            "The `filename` in the `add_to_package` decorator hook requires a unique "
            "`file_path` to `file_name` mapping" % (filename, ", ".join(file_paths))
        )
        super().__init__(msg=msg, lineno=lineno)


class MetaflowPackage(object):
    def __init__(
        self,
        flow,
        environment,
        echo,
        suffixes: Optional[List[str]] = DEFAULT_SUFFIXES_LIST,
        user_code_filter: Optional[Callable[[str], bool]] = None,
        flow_datastore: Optional["metaflow.datastore.FlowDataStore"] = None,
        mfcontent: Optional[MetaflowCodeContent] = None,
        exclude_tl_dirs=None,
        backend: Type[PackagingBackend] = TarPackagingBackend,
    ):
        self._environment = environment
        self._environment.init_environment(echo)

        self._echo = echo
        self._flow = flow
        self._flow_datastore = flow_datastore
        self._backend = backend

        # Info about the package
        self._name = None
        self._create_time = time.time()
        self._user_flow_dir = None

        # Content of the package (and settings on how to create it)
        if suffixes is not None:
            self._suffixes = list(set().union(suffixes, DEFAULT_SUFFIXES_LIST))
        else:
            self._suffixes = None

        def _module_selector(m) -> bool:
            from ..user_decorators.user_flow_decorator import FlowMutatorMeta
            from ..user_decorators.user_step_decorator import UserStepDecoratorMeta

            # Be very defensive here to filter modules in case there are
            # some badly behaved modules that have weird values for
            # METAFLOW_PACKAGE_POLICY for example.
            try:
                if (
                    m.__name__ in FlowMutatorMeta._import_modules
                    or m.__name__ in UserStepDecoratorMeta._import_modules
                    or (
                        hasattr(m, "METAFLOW_PACKAGE_POLICY")
                        and m.METAFLOW_PACKAGE_POLICY == "include"
                    )
                ):
                    return True
                return False
            except:
                return False

        if mfcontent is None:
            self._mfcontent = MetaflowCodeContentV1(criteria=_module_selector)

        else:
            self._mfcontent = mfcontent
        # We exclude the environment when packaging as this will be packaged separately.
        # This comes into play primarily if packaging from a node already running packaged
        # code.
        # These directories are only excluded at the top-level (ie: not further down
        # in sub-directories)
        # "_escape_trampolines" is a special directory where trampoline escape hatch
        # files are stored (used by Netflix Extension's Conda implementation).
        self._exclude_tl_dirs = (
            self._mfcontent.get_excluded_tl_entries()
            + ["_escape_trampolines"]
            + (exclude_tl_dirs or [])
        )

        if self._suffixes is not None and user_code_filter is not None:
            self._user_code_filter = lambda x, f1=suffix_filter(
                self._suffixes
            ), f2=user_code_filter: f1(x) and f2(x)
            self._filter_type = "suffixes and user filter"
        elif self._suffixes is not None:
            self._user_code_filter = suffix_filter(self._suffixes)
            self._filter_type = "suffixes"
        elif user_code_filter is not None:
            self._user_code_filter = user_code_filter
            self._filter_type = "user filter"
        else:
            self._user_code_filter = lambda x: True
            self._filter_type = "no filter"

        # Info about the package creation (it happens async)
        self._is_package_available = None
        self._blob_sha = None
        self._blob_url = None
        self._blob = None

        # We launch a thread to create the package asynchronously and upload
        # it opportunistically
        self._create_thread = threading.Thread(
            target=self._package_and_upload,
            daemon=True,
        )
        self._create_thread.start()

    # HORRIBLE HACK SO THAT CURRENT COMPUTE IMPLEMENTATIONS CAN STILL
    # DO pkg.blob. Ideally, this goes away and blob_with_timeout becomes
    # the main method (called blob).
    @property
    def blob(self) -> BytesIO:
        return self.blob_with_timeout()

    def blob_with_timeout(self, timeout: Optional[float] = None) -> BytesIO:
        if self._blob is None:
            self._create_thread.join(timeout)
            if self._is_package_available is not None:
                # We have our result now
                if self._is_package_available:
                    return self._blob
                else:
                    raise self._packaging_exception
        return self._blob

    def package_sha(self, timeout: Optional[float] = None) -> Optional[str]:
        if self._blob_sha is None:
            self._create_thread.join(timeout)
            if self._is_package_available is not None:
                # We have our result now
                if self._is_package_available:
                    return self._blob_sha
                else:
                    raise self._packaging_exception
        return self._blob_sha

    def package_url(self, timeout: Optional[float] = None) -> Optional[str]:
        if self._blob_url is None:
            self._create_thread.join(timeout)
            if self._is_package_available is not None:
                # We have our result now
                if self._is_package_available:
                    return self._blob_url
                else:
                    raise self._packaging_exception
        return self._blob_url

    @property
    def package_metadata(self):
        return json.dumps(
            {
                "version": 0,
                "archive_format": self._backend.backend_type(),
                "mfcontent_version": self._mfcontent.get_package_version(),
            }
        )

    @classmethod
    def get_backend(cls, pkg_metadata: str) -> PackagingBackend:
        """
        Method to get the backend type from the package metadata.

        Parameters
        ----------
        pkg_metadata : str
            The metadata of the package to extract.

        Returns
        -------
        PackagingBackend
            The backend type that can be used to extract the package.
        """
        backend_type = json.loads(pkg_metadata).get("archive_format", "tgz")
        return PackagingBackend.get_backend(backend_type)

    @classmethod
    def get_extract_commands(
        cls, pkg_metadata: str, archive_path: str, dest_dir: str = "."
    ) -> List[str]:
        """
        Method to get the commands needed to extract the package into
        the directory dest_dir. Note that this will return a list of commands
        that can be passed to subprocess.run for example.

        Parameters
        ----------
        pkg_metadata : str
            The metadata of the package to extract.
        archive_path : str
            The path to the archive to extract.
        dest_dir : str, default "."
            The directory to extract the package into.

        Returns
        -------
        List[str]
            The commands needed to extract the package into the directory dest_dir.
        """
        backend_type = json.loads(pkg_metadata).get("archive_format", "tgz")
        # We now ask the backend type how to extract itself
        backend = PackagingBackend.get_backend(backend_type)
        cmds = backend.get_extract_commands(archive_path, dest_dir)
        debug.package_exec(f"Command to extract {archive_path} into {dest_dir}: {cmds}")
        return cmds

    @classmethod
    def get_post_extract_env_vars(
        cls, pkg_metadata: str, dest_dir: str = "."
    ) -> Dict[str, str]:
        """
        Method to get the environment variables needed to access the content
        that has been extracted into the directory dest_dir. This will
        typically involve setting PYTHONPATH

        Parameters
        ----------
        pkg_metadata : str
            The metadata of the package to extract.
        dest_dir : str, default "."
            The directory where the content has been extracted to.

        Returns
        -------
        Dict[str, str]
            The post-extract environment variables that are needed to access the content
            that has been extracted into dest_dir.
        """
        mfcontent_version = json.loads(pkg_metadata).get("mfcontent_version", 0)
        env_vars = MetaflowCodeContent.get_post_extract_env_vars(
            mfcontent_version, dest_dir
        )
        debug.package_exec(
            f"Environment variables to access content extracted into {dest_dir}: {env_vars}"
        )
        return env_vars

    @classmethod
    def cls_get_content(
        cls, pkg_metadata, archive: BytesIO, name: str
    ) -> Optional[bytes]:
        """
        Method to get the content of a member in the package archive.

        Parameters
        ----------
        pkg_metadata : str
            The metadata of the package to extract.
        archive : BytesIO
            The archive to extract the member from.
        name : str
            The name of the member to extract.

        Returns
        -------
        Optional[bytes]
            The content of the member if it exists, None otherwise.
        """
        backend = cls.get_backend(pkg_metadata)
        with backend.cls_open(archive) as opened_archive:
            return backend.cls_get_member(opened_archive, name)

    @classmethod
    def cls_get_info(cls, pkg_metadata, archive: BytesIO) -> Optional[Dict[str, str]]:
        """
        Method to get the info of the package from the archive.
        Parameters
        ----------
        pkg_metadata : str
            The metadata of the package to extract.
        archive : BytesIO
            The archive to extract the info from.
        Returns
        -------
        Optional[Dict[str, str]]
            The info of the package if it exists, None otherwise.
        """
        backend = cls.get_backend(pkg_metadata)
        with backend.cls_open(archive) as opened_archive:
            return MetaflowCodeContent.get_archive_info(opened_archive, backend)

    @classmethod
    def cls_get_config(
        cls, pkg_metadata: str, archive: BytesIO
    ) -> Optional[Dict[str, str]]:
        """
        Method to get the config of the package from the archive.

        Parameters
        ----------
        pkg_metadata : str
            The metadata of the package to extract.
        archive : BytesIO
            The archive to extract the config from.

        Returns
        -------
        Optional[Dict[str, str]]
            The config of the package if it exists, None otherwise.
        """
        backend = cls.get_backend(pkg_metadata)
        with backend.cls_open(archive) as opened_archive:
            return MetaflowCodeContent.get_archive_config(opened_archive, backend)

    @classmethod
    def cls_extract_into(
        cls,
        pkg_metadata: str,
        archive: BytesIO,
        dest_dir: str = ".",
        content_types: int = ContentType.ALL_CONTENT.value,
    ):
        """
        Method to extract the package archive into a directory.

        Parameters
        ----------
        pkg_metadata : str
            The metadata of the package to extract.
        archive : BytesIO
            The archive to extract.
        dest_dir : str, default "."
            The directory to extract the package into.
        content_types : int, default ALL_CONTENT
            The types of content to extract. This is a bitmask of ContentType values.
        """
        backend = cls.get_backend(pkg_metadata)
        with backend.cls_open(archive) as opened_archive:
            include_members = MetaflowCodeContent.get_archive_content_members(
                opened_archive, content_types, backend
            )
            backend.cls_extract_members(opened_archive, include_members, dest_dir)

    def user_tuples(self, timeout: Optional[float] = None):
        # Wait for at least the blob to be formed
        _ = self.blob_with_timeout(timeout=timeout)
        for path, arcname in self._cached_user_members:
            yield path, arcname

    def path_tuples(self, timeout: Optional[float] = None):
        # Wait for at least the blob to be formed
        _ = self.blob_with_timeout(timeout=timeout)
        # Files included in the environment
        yield from self._mfcontent.content_names()

        # Files included in the user code
        yield from self.user_tuples()

    def show(self, timeout: Optional[float] = None) -> str:
        # Human-readable content of the package
        blob = self.blob_with_timeout(timeout=timeout)  # Ensure the package is created
        lines = [
            f"Package size: {self._format_size(len(blob))}",
            f"Number of files: {sum(1 for _ in self.path_tuples())}",
            self._mfcontent.show(),
        ]

        if self._flow:
            lines.append(f"\nUser code in flow {self._name}:")
            lines.append(f"  - Packaged from directory {self._user_flow_dir}")
            if self._filter_type != "no filter":
                if self._suffixes:
                    lines.append(
                        f"  - Filtered by suffixes: {', '.join(self._suffixes)}"
                    )
                else:
                    lines.append(f"  - Filtered by {self._filter_type}")
            else:
                lines.append("  - No user code filter applied")
            if self._exclude_tl_dirs:
                lines.append(
                    f"  - Excluded directories: {', '.join(self._exclude_tl_dirs)}"
                )
        return "\n".join(lines)

    def get_content(
        self, name: str, content_type: ContentType, timeout: Optional[float] = None
    ) -> Optional[bytes]:
        """
        Method to get the content of a file within the package. This method
        should be used for one-off access to small-ish files. If more files are
        needed, use extract_into to extract the package into a directory and
        then access the files from there.

        Parameters
        ----------
        name : str
            The name of the file to get the content of. Note that this
            is not necessarily the name in the archive but is the name
            that was passed in when creating the archive (in the archive,
            it may be prefixed by some directory structure).
        content_type : ContentType
            The type of file to get the content of.

        Returns
        -------
        Optional[bytes]
            The content of the file. If the file is not found, None is returned.
        """
        # Wait for at least the blob to be formed
        _ = self.blob_with_timeout(timeout=timeout)
        if content_type == ContentType.USER_CONTENT:
            for path, arcname in self.user_tuples():
                if name == arcname:
                    return open(path, "rb").read()
            return None
        elif content_type in (
            ContentType.CODE_CONTENT,
            ContentType.MODULE_CONTENT,
            ContentType.OTHER_CONTENT,
        ):
            mangled_name = self._mfcontent.get_archive_filename(name, content_type)
            for path_or_bytes, arcname in self._mfcontent.contents(content_type):
                if mangled_name == arcname:
                    if isinstance(path_or_bytes, bytes):
                        # In case this is generated content like an INFO file
                        return path_or_bytes
                    # Otherwise, it is a file path
                    return open(path_or_bytes, "rb").read()
            return None
        raise ValueError(f"Unknown content type: {content_type}")

    def extract_into(
        self,
        dest_dir: str = ".",
        content_types: int = ContentType.ALL_CONTENT.value,
        timeout: Optional[float] = None,
    ):
        """
        Method to extract the package (or some of the files) into a directory.

        Parameters
        ----------
        dest_dir : str, default "."
            The directory to extract the package into.
        content_types : int, default ALL_CONTENT
            The types of content to extract.
        """
        _ = self.blob_with_timeout(timeout=timeout)  # Ensure the package is created
        member_list = []
        if content_types & ContentType.USER_CONTENT.value:
            member_list.extend(
                [(m[0], os.path.join(dest_dir, m[1])) for m in self.user_tuples()]
            )
        if content_types & (
            ContentType.CODE_CONTENT.value | ContentType.MODULE_CONTENT.value
        ):
            # We need to get the name of the files in the content archive to extract
            member_list.extend(
                [
                    (m[0], os.path.join(dest_dir, m[1]))
                    for m in self._mfcontent.content_names(
                        content_types & ~ContentType.OTHER_CONTENT.value
                    )
                ]
            )
        for orig_path, new_path in member_list:
            os.makedirs(os.path.dirname(new_path), exist_ok=True)
            # TODO: In case there are duplicate files -- that should not be the case
            # but there is a bug currently with internal Netflix code.
            if not os.path.exists(new_path):
                os.symlink(orig_path, new_path)
            # Could copy files as well if we want to split them out.
            # shutil.copy(orig_path, new_path)
        # OTHER_CONTENT requires special handling because sometimes the file isn't a file
        # but generated content
        member_list = []
        if content_types & ContentType.OTHER_CONTENT.value:
            member_list.extend(
                [
                    (m[0], os.path.join(dest_dir, m[1]))
                    for m in self._mfcontent.contents(ContentType.OTHER_CONTENT)
                ]
            )
        for path_or_content, new_path in member_list:
            os.makedirs(os.path.dirname(new_path), exist_ok=True)
            if not os.path.exists(new_path):
                if isinstance(path_or_content, bytes):
                    with open(new_path, "wb") as f:
                        f.write(path_or_content)
                else:
                    os.symlink(path_or_content, new_path)

    @staticmethod
    def _format_size(size_in_bytes):
        for unit in ["B", "KB", "MB", "GB", "TB"]:
            if size_in_bytes < 1024.0:
                return f"{size_in_bytes:.2f} {unit}"
            size_in_bytes /= 1024.0
        return f"{size_in_bytes:.2f} PB"

    def _package_and_upload(self):
        try:
            # Can be called without a flow (Function)
            if self._flow:
                for step in self._flow:
                    for deco in step.decorators:
                        deco.package_init(self._flow, step.__name__, self._environment)
                self._name = f"flow {self._flow.name}"
            else:
                self._name = "<generic code package>"

            # Add metacontent
            self._mfcontent.add_info(
                self._environment.get_environment_info(include_ext_info=True)
            )

            self._mfcontent.add_config(dump_config_values(self._flow))

            # Add user files (from decorators and environment)
            if self._flow:
                self._add_addl_files()
                self._cached_user_members = list(self._user_code_tuples())
                debug.package_exec(
                    f"User files to package: {self._cached_user_members}"
                )

            self._blob = self._make()
            if self._flow_datastore:
                if len(self._blob) > 100 * 1024 * 1024:
                    self._echo(
                        f"Warning: The code package for {self._flow.name} is larger than "
                        f"100MB (found it to be {self._format_size(len(self._blob))}) "
                        "This may lead to slower upload times for remote runs and no "
                        "uploads for local runs. Consider reducing the package size. "
                        "Use `<myflow.py> package info` or `<myflow.py> package list` "
                        "to get more information about what is included in the package."
                    )
                self._blob_url, self._blob_sha = self._flow_datastore.save_data(
                    [self._blob], len_hint=1
                )[0]
            else:
                self._blob_url = self._blob_sha = ""
            self._is_package_available = True
        except Exception as e:
            self._packaging_exception = e
            self._echo(f"Package creation/upload failed for {self._flow.name}: {e}")
            self._is_package_available = False

    def _add_addl_files(self):
        # Look at all decorators that provide additional files
        deco_module_paths = {}
        addl_modules = set()

        def _check_tuple(path_tuple):
            if len(path_tuple) == 2:
                path_tuple = (
                    path_tuple[0],
                    path_tuple[1],
                    ContentType.CODE_CONTENT,
                )
            file_path, file_name, file_type = path_tuple
            if file_type == ContentType.MODULE_CONTENT:
                if file_path in addl_modules:
                    return None  # Module was already added -- we don't add twice
                addl_modules.add(file_path)
            elif file_type in (
                ContentType.OTHER_CONTENT,
                ContentType.CODE_CONTENT,
            ):
                path_tuple = (os.path.realpath(path_tuple[0]), path_tuple[1], file_type)
                # These are files
                # Check if the path is not duplicated as
                # many steps can have the same packages being imported
                if file_name not in deco_module_paths:
                    deco_module_paths[file_name] = file_path
                elif deco_module_paths[file_name] != file_path:
                    raise NonUniqueFileNameToFilePathMappingException(
                        file_name, [deco_module_paths[file_name], file_path]
                    )
            else:
                raise ValueError(f"Unknown file type: {file_type}")
            return path_tuple

        def _add_tuple(path_tuple):
            file_path, file_name, file_type = path_tuple
            if file_type == ContentType.MODULE_CONTENT:
                # file_path is actually a module
                self._mfcontent.add_module(cast(ModuleType, file_path))
            elif file_type == ContentType.CODE_CONTENT:
                self._mfcontent.add_code_file(file_path, file_name)
            elif file_type == ContentType.OTHER_CONTENT:
                self._mfcontent.add_other_file(file_path, file_name)

        for step in self._flow:
            for deco in step.decorators:
                for path_tuple in deco.add_to_package():
                    path_tuple = _check_tuple(path_tuple)
                    if path_tuple is None:
                        continue
                    _add_tuple(path_tuple)

        # the package folders for environment
        for path_tuple in self._environment.add_to_package():
            path_tuple = _check_tuple(path_tuple)
            if path_tuple is None:
                continue
            _add_tuple(path_tuple)

    def _user_code_tuples(self):
        if R.use_r():
            # the R working directory
            self._user_flow_dir = R.working_dir()
            for path_tuple in walk(
                "%s/" % R.working_dir(), file_filter=self._user_code_filter
            ):
                yield path_tuple
            # the R package
            for path_tuple in R.package_paths():
                yield path_tuple
        else:
            # the user's working directory
            flowdir = os.path.dirname(os.path.abspath(sys.argv[0])) + "/"
            self._user_flow_dir = flowdir
            for path_tuple in walk(
                flowdir,
                file_filter=self._user_code_filter,
                exclude_tl_dirs=self._exclude_tl_dirs,
            ):
                # TODO: This is where we will check if the file is already included
                # in the mfcontent portion
                yield path_tuple

    def _make(self):
        backend = self._backend()
        with backend.create() as archive:
            # Package the environment
            for path_or_bytes, arcname in self._mfcontent.contents():
                if isinstance(path_or_bytes, str):
                    archive.add_file(path_or_bytes, arcname=arcname)
                else:
                    archive.add_data(BytesIO(path_or_bytes), arcname=arcname)

            # Package the user code
            for path, arcname in self._cached_user_members:
                archive.add_file(path, arcname=arcname)
        return backend.get_blob()

    def __str__(self):
        return f"<code package for {self._name} (created @ {self._create_time})>"



import json
import math
import os
import shlex
import time
from uuid import uuid4

from metaflow import current, util
from metaflow.exception import MetaflowException
from metaflow.metaflow_config import (
    ARGO_EVENTS_EVENT,
    ARGO_EVENTS_EVENT_BUS,
    ARGO_EVENTS_EVENT_SOURCE,
    ARGO_EVENTS_INTERNAL_WEBHOOK_URL,
    ARGO_EVENTS_SERVICE_ACCOUNT,
    ARGO_EVENTS_WEBHOOK_AUTH,
    ARGO_WORKFLOWS_KUBERNETES_SECRETS,
    ARGO_WORKFLOWS_ENV_VARS_TO_SKIP,
    AWS_SECRETS_MANAGER_DEFAULT_REGION,
    AZURE_KEY_VAULT_PREFIX,
    AZURE_STORAGE_BLOB_SERVICE_ENDPOINT,
    CARD_AZUREROOT,
    CARD_GSROOT,
    CARD_S3ROOT,
    DATASTORE_SYSROOT_AZURE,
    DATASTORE_SYSROOT_GS,
    DATASTORE_SYSROOT_S3,
    DATATOOLS_S3ROOT,
    DEFAULT_AWS_CLIENT_PROVIDER,
    DEFAULT_GCP_CLIENT_PROVIDER,
    DEFAULT_METADATA,
    DEFAULT_SECRETS_BACKEND_TYPE,
    GCP_SECRET_MANAGER_PREFIX,
    KUBERNETES_FETCH_EC2_METADATA,
    KUBERNETES_SANDBOX_INIT_SCRIPT,
    OTEL_ENDPOINT,
    S3_ENDPOINT_URL,
    S3_SERVER_SIDE_ENCRYPTION,
    SERVICE_HEADERS,
    KUBERNETES_SECRETS,
    SERVICE_INTERNAL_URL,
)
from metaflow.unbounded_foreach import UBF_CONTROL, UBF_TASK
from metaflow.metaflow_config_funcs import config_values, init_config
from metaflow.mflog import (
    BASH_SAVE_LOGS,
    bash_capture_logs,
    export_mflog_env_vars,
    get_log_tailer,
    tail_logs,
)

from .kubernetes_client import KubernetesClient

# Redirect structured logs to $PWD/.logs/
LOGS_DIR = "$PWD/.logs"
STDOUT_FILE = "mflog_stdout"
STDERR_FILE = "mflog_stderr"
STDOUT_PATH = os.path.join(LOGS_DIR, STDOUT_FILE)
STDERR_PATH = os.path.join(LOGS_DIR, STDERR_FILE)

METAFLOW_PARALLEL_STEP_CLI_OPTIONS_TEMPLATE = (
    "{METAFLOW_PARALLEL_STEP_CLI_OPTIONS_TEMPLATE}"
)


class KubernetesException(MetaflowException):
    headline = "Kubernetes error"


class KubernetesKilledException(MetaflowException):
    headline = "Kubernetes Batch job killed"


class Kubernetes(object):
    def __init__(
        self,
        datastore,
        metadata,
        environment,
    ):
        self._datastore = datastore
        self._metadata = metadata
        self._environment = environment

    def _command(
        self,
        flow_name,
        run_id,
        step_name,
        task_id,
        attempt,
        code_package_metadata,
        code_package_url,
        step_cmds,
    ):
        mflog_expr = export_mflog_env_vars(
            flow_name=flow_name,
            run_id=run_id,
            step_name=step_name,
            task_id=task_id,
            retry_count=attempt,
            datastore_type=self._datastore.TYPE,
            stdout_path=STDOUT_PATH,
            stderr_path=STDERR_PATH,
        )
        init_cmds = self._environment.get_package_commands(
            code_package_url, self._datastore.TYPE, code_package_metadata
        )
        init_expr = " && ".join(init_cmds)
        step_expr = bash_capture_logs(
            " && ".join(
                self._environment.bootstrap_commands(step_name, self._datastore.TYPE)
                + step_cmds
            )
        )

        # Construct an entry point that
        # 1) initializes the mflog environment (mflog_expr)
        # 2) bootstraps a metaflow environment (init_expr)
        # 3) executes a task (step_expr)

        # The `true` command is to make sure that the generated command
        # plays well with docker containers which have entrypoint set as
        # eval $@
        cmd_str = "true && mkdir -p %s && %s && %s && %s; " % (
            LOGS_DIR,
            mflog_expr,
            init_expr,
            step_expr,
        )
        # After the task has finished, we save its exit code (fail/success)
        # and persist the final logs. The whole entrypoint should exit
        # with the exit code (c) of the task.
        #
        # Note that if step_expr OOMs, this tail expression is never executed.
        # We lose the last logs in this scenario.
        #
        # TODO: Capture hard exit logs in Kubernetes.
        cmd_str += "c=$?; %s; exit $c" % BASH_SAVE_LOGS
        # For supporting sandboxes, ensure that a custom script is executed before
        # anything else is executed. The script is passed in as an env var.
        cmd_str = (
            '${METAFLOW_INIT_SCRIPT:+eval \\"${METAFLOW_INIT_SCRIPT}\\"} && %s'
            % cmd_str
        )
        return shlex.split('bash -c "%s"' % cmd_str)

    def launch_job(self, **kwargs):
        if (
            "num_parallel" in kwargs
            and kwargs["num_parallel"]
            and int(kwargs["num_parallel"]) > 0
        ):
            self._job = self.create_jobset(**kwargs).execute()
        else:
            kwargs.pop("num_parallel", None)
            kwargs["name_pattern"] = "t-{uid}-".format(uid=str(uuid4())[:8])
            self._job = self.create_job_object(**kwargs).create().execute()

    def create_jobset(
        self,
        flow_name,
        run_id,
        step_name,
        task_id,
        attempt,
        user,
        code_package_metadata,
        code_package_sha,
        code_package_url,
        code_package_ds,
        docker_image,
        docker_image_pull_policy,
        image_pull_secrets=None,
        step_cli=None,
        service_account=None,
        secrets=None,
        node_selector=None,
        namespace=None,
        cpu=None,
        gpu=None,
        gpu_vendor=None,
        disk=None,
        memory=None,
        use_tmpfs=None,
        tmpfs_tempdir=None,
        tmpfs_size=None,
        tmpfs_path=None,
        run_time_limit=None,
        env=None,
        persistent_volume_claims=None,
        tolerations=None,
        labels=None,
        annotations=None,
        shared_memory=None,
        port=None,
        num_parallel=None,
        qos=None,
        security_context=None,
    ):
        name = "js-%s" % str(uuid4())[:6]
        jobset = (
            KubernetesClient()
            .jobset(
                name=name,
                namespace=namespace,
                service_account=service_account,
                node_selector=node_selector,
                image=docker_image,
                image_pull_policy=docker_image_pull_policy,
                image_pull_secrets=image_pull_secrets,
                cpu=cpu,
                memory=memory,
                disk=disk,
                gpu=gpu,
                gpu_vendor=gpu_vendor,
                timeout_in_seconds=run_time_limit,
                # Retries are handled by Metaflow runtime
                retries=0,
                step_name=step_name,
                # We set the jobset name as the subdomain.
                # todo: [final-refactor] ask @shri what was the motive when we did initial implementation
                subdomain=name,
                tolerations=tolerations,
                use_tmpfs=use_tmpfs,
                tmpfs_tempdir=tmpfs_tempdir,
                tmpfs_size=tmpfs_size,
                tmpfs_path=tmpfs_path,
                persistent_volume_claims=persistent_volume_claims,
                shared_memory=shared_memory,
                port=port,
                num_parallel=num_parallel,
                qos=qos,
                security_context=security_context,
            )
            .environment_variable("METAFLOW_CODE_METADATA", code_package_metadata)
            .environment_variable("METAFLOW_CODE_SHA", code_package_sha)
            .environment_variable("METAFLOW_CODE_URL", code_package_url)
            .environment_variable("METAFLOW_CODE_DS", code_package_ds)
            .environment_variable("METAFLOW_USER", user)
            .environment_variable("METAFLOW_SERVICE_URL", SERVICE_INTERNAL_URL)
            .environment_variable(
                "METAFLOW_SERVICE_HEADERS",
                json.dumps(SERVICE_HEADERS),
            )
            .environment_variable("METAFLOW_DATASTORE_SYSROOT_S3", DATASTORE_SYSROOT_S3)
            .environment_variable("METAFLOW_DATATOOLS_S3ROOT", DATATOOLS_S3ROOT)
            .environment_variable("METAFLOW_DEFAULT_DATASTORE", self._datastore.TYPE)
            .environment_variable("METAFLOW_DEFAULT_METADATA", DEFAULT_METADATA)
            .environment_variable("METAFLOW_KUBERNETES_WORKLOAD", 1)
            .environment_variable(
                "METAFLOW_KUBERNETES_FETCH_EC2_METADATA", KUBERNETES_FETCH_EC2_METADATA
            )
            .environment_variable("METAFLOW_RUNTIME_ENVIRONMENT", "kubernetes")
            .environment_variable(
                "METAFLOW_DEFAULT_SECRETS_BACKEND_TYPE", DEFAULT_SECRETS_BACKEND_TYPE
            )
            .environment_variable("METAFLOW_CARD_S3ROOT", CARD_S3ROOT)
            .environment_variable(
                "METAFLOW_DEFAULT_AWS_CLIENT_PROVIDER", DEFAULT_AWS_CLIENT_PROVIDER
            )
            .environment_variable(
                "METAFLOW_DEFAULT_GCP_CLIENT_PROVIDER", DEFAULT_GCP_CLIENT_PROVIDER
            )
            .environment_variable(
                "METAFLOW_AWS_SECRETS_MANAGER_DEFAULT_REGION",
                AWS_SECRETS_MANAGER_DEFAULT_REGION,
            )
            .environment_variable(
                "METAFLOW_GCP_SECRET_MANAGER_PREFIX", GCP_SECRET_MANAGER_PREFIX
            )
            .environment_variable(
                "METAFLOW_AZURE_KEY_VAULT_PREFIX", AZURE_KEY_VAULT_PREFIX
            )
            .environment_variable("METAFLOW_S3_ENDPOINT_URL", S3_ENDPOINT_URL)
            .environment_variable(
                "METAFLOW_AZURE_STORAGE_BLOB_SERVICE_ENDPOINT",
                AZURE_STORAGE_BLOB_SERVICE_ENDPOINT,
            )
            .environment_variable(
                "METAFLOW_DATASTORE_SYSROOT_AZURE", DATASTORE_SYSROOT_AZURE
            )
            .environment_variable("METAFLOW_CARD_AZUREROOT", CARD_AZUREROOT)
            .environment_variable("METAFLOW_DATASTORE_SYSROOT_GS", DATASTORE_SYSROOT_GS)
            .environment_variable("METAFLOW_CARD_GSROOT", CARD_GSROOT)
            # support Metaflow sandboxes
            .environment_variable(
                "METAFLOW_INIT_SCRIPT", KUBERNETES_SANDBOX_INIT_SCRIPT
            )
            .environment_variable(
                "METAFLOW_KUBERNETES_SANDBOX_INIT_SCRIPT",
                KUBERNETES_SANDBOX_INIT_SCRIPT,
            )
            .environment_variable(
                "METAFLOW_ARGO_WORKFLOWS_KUBERNETES_SECRETS",
                ARGO_WORKFLOWS_KUBERNETES_SECRETS,
            )
            .environment_variable(
                "METAFLOW_ARGO_WORKFLOWS_ENV_VARS_TO_SKIP",
                ARGO_WORKFLOWS_ENV_VARS_TO_SKIP,
            )
            .environment_variable("METAFLOW_OTEL_ENDPOINT", OTEL_ENDPOINT)
            # Skip setting METAFLOW_DATASTORE_SYSROOT_LOCAL because metadata sync
            # between the local user instance and the remote Kubernetes pod
            # assumes metadata is stored in DATASTORE_LOCAL_DIR on the Kubernetes
            # pod; this happens when METAFLOW_DATASTORE_SYSROOT_LOCAL is NOT set (
            # see get_datastore_root_from_config in datastore/local.py).
        )

        for k in list(
            [] if not secrets else [secrets] if isinstance(secrets, str) else secrets
        ) + KUBERNETES_SECRETS.split(","):
            jobset.secret(k)

        initial_configs = init_config()
        for entry in ["OBP_PERIMETER", "OBP_INTEGRATIONS_URL"]:
            if entry not in initial_configs:
                raise KubernetesException(
                    f"{entry} was not found in metaflow config. Please make sure to run `outerbounds configure <...>` command which can be found on the Ourebounds UI or reach out to your Outerbounds support team."
                )

        additional_obp_configs = {
            "OBP_PERIMETER": initial_configs["OBP_PERIMETER"],
            "OBP_INTEGRATIONS_URL": initial_configs[
                "OBP_INTEGRATIONS_URL"
            ],
        }
        for k, v in additional_obp_configs.items():
            jobset.environment_variable(k, v)

        jobset.environment_variables_from_selectors(
            {
                "METAFLOW_KUBERNETES_NAMESPACE": "metadata.namespace",
                "METAFLOW_KUBERNETES_POD_NAMESPACE": "metadata.namespace",
                "METAFLOW_KUBERNETES_POD_NAME": "metadata.name",
                "METAFLOW_KUBERNETES_POD_ID": "metadata.uid",
                "METAFLOW_KUBERNETES_SERVICE_ACCOUNT_NAME": "spec.serviceAccountName",
                "METAFLOW_KUBERNETES_NODE_IP": "status.hostIP",
            }
        )

        # Temporary passing of *some* environment variables. Do not rely on this
        # mechanism as it will be removed in the near future
        for k, v in config_values():
            if k.startswith("METAFLOW_CONDA_") or k.startswith("METAFLOW_DEBUG_"):
                jobset.environment_variable(k, v)

        if S3_SERVER_SIDE_ENCRYPTION is not None:
            jobset.environment_variable(
                "METAFLOW_S3_SERVER_SIDE_ENCRYPTION", S3_SERVER_SIDE_ENCRYPTION
            )

        # Set environment variables to support metaflow.integrations.ArgoEvent
        jobset.environment_variable(
            "METAFLOW_ARGO_EVENTS_WEBHOOK_URL", ARGO_EVENTS_INTERNAL_WEBHOOK_URL
        )
        jobset.environment_variable("METAFLOW_ARGO_EVENTS_EVENT", ARGO_EVENTS_EVENT)
        jobset.environment_variable(
            "METAFLOW_ARGO_EVENTS_EVENT_BUS", ARGO_EVENTS_EVENT_BUS
        )
        jobset.environment_variable(
            "METAFLOW_ARGO_EVENTS_EVENT_SOURCE", ARGO_EVENTS_EVENT_SOURCE
        )
        jobset.environment_variable(
            "METAFLOW_ARGO_EVENTS_SERVICE_ACCOUNT", ARGO_EVENTS_SERVICE_ACCOUNT
        )
        jobset.environment_variable(
            "METAFLOW_ARGO_EVENTS_WEBHOOK_AUTH",
            ARGO_EVENTS_WEBHOOK_AUTH,
        )

        ## -----Jobset specific env vars START here-----
        jobset.environment_variable("MF_MASTER_ADDR", jobset.jobset_control_addr)
        jobset.environment_variable("MF_MASTER_PORT", str(port))
        jobset.environment_variable("MF_WORLD_SIZE", str(num_parallel))
        jobset.environment_variable_from_selector(
            "JOBSET_RESTART_ATTEMPT",
            "metadata.annotations['jobset.sigs.k8s.io/restart-attempt']",
        )
        jobset.environment_variable_from_selector(
            "METAFLOW_KUBERNETES_JOBSET_NAME",
            "metadata.annotations['jobset.sigs.k8s.io/jobset-name']",
        )
        jobset.environment_variable_from_selector(
            "MF_WORKER_REPLICA_INDEX",
            "metadata.annotations['jobset.sigs.k8s.io/job-index']",
        )
        ## -----Jobset specific env vars END here-----

        tmpfs_enabled = use_tmpfs or (tmpfs_size and not use_tmpfs)
        if tmpfs_enabled and tmpfs_tempdir:
            jobset.environment_variable("METAFLOW_TEMPDIR", tmpfs_path)

        for name, value in env.items():
            jobset.environment_variable(name, value)

        system_annotations = {
            "metaflow/user": user,
            "metaflow/flow_name": flow_name,
            "metaflow/control-task-id": task_id,
            "metaflow/run_id": run_id,
            "metaflow/step_name": step_name,
            "metaflow/attempt": attempt,
        }
        if current.get("project_name"):
            system_annotations.update(
                {
                    "metaflow/project_name": current.project_name,
                    "metaflow/branch_name": current.branch_name,
                    "metaflow/project_flow_name": current.project_flow_name,
                }
            )

        system_labels = {
            "app.kubernetes.io/name": "metaflow-task",
            "app.kubernetes.io/part-of": "metaflow",
        }

        jobset.labels({**({} if not labels else labels), **system_labels})

        jobset.annotations(
            {**({} if not annotations else annotations), **system_annotations}
        )
        # We need this task-id set so that all the nodes are aware of the control
        # task's task-id. These "MF_" variables populate the `current.parallel` namedtuple
        jobset.environment_variable("MF_PARALLEL_CONTROL_TASK_ID", str(task_id))

        ## ----------- control/worker specific values START here -----------
        # We will now set the appropriate command for the control/worker job
        _get_command = lambda index, _tskid: self._command(
            flow_name=flow_name,
            run_id=run_id,
            step_name=step_name,
            task_id=_tskid,
            attempt=attempt,
            code_package_metadata=code_package_metadata,
            code_package_url=code_package_url,
            step_cmds=[
                step_cli.replace(
                    METAFLOW_PARALLEL_STEP_CLI_OPTIONS_TEMPLATE,
                    "--ubf-context $UBF_CONTEXT --split-index %s --task-id %s"
                    % (index, _tskid),
                )
            ],
        )
        jobset.control.replicas(1)
        jobset.worker.replicas(num_parallel - 1)

        # We set the appropriate command for the control/worker job
        # and also set the task-id/spit-index for the control/worker job
        # appropirately.
        jobset.control.command(_get_command("0", str(task_id)))
        jobset.worker.command(
            _get_command(
                "`expr $[MF_WORKER_REPLICA_INDEX] + 1`",
                "-".join(
                    [
                        str(task_id),
                        "worker",
                        "$MF_WORKER_REPLICA_INDEX",
                    ]
                ),
            )
        )

        jobset.control.environment_variable("UBF_CONTEXT", UBF_CONTROL)
        jobset.worker.environment_variable("UBF_CONTEXT", UBF_TASK)
        # Every control job requires an environment variable of MF_CONTROL_INDEX
        # set to 0 so that we can derive the MF_PARALLEL_NODE_INDEX correctly.
        # Since only the control job has MF_CONTROL_INDE set to 0, all worker nodes
        # will use MF_WORKER_REPLICA_INDEX
        jobset.control.environment_variable("MF_CONTROL_INDEX", "0")
        ## ----------- control/worker specific values END here -----------

        return jobset

    def create_job_object(
        self,
        flow_name,
        run_id,
        step_name,
        task_id,
        attempt,
        user,
        code_package_metadata,
        code_package_sha,
        code_package_url,
        code_package_ds,
        step_cli,
        docker_image,
        docker_image_pull_policy,
        image_pull_secrets=None,
        service_account=None,
        secrets=None,
        node_selector=None,
        namespace=None,
        cpu=None,
        gpu=None,
        gpu_vendor=None,
        disk=None,
        memory=None,
        use_tmpfs=None,
        tmpfs_tempdir=None,
        tmpfs_size=None,
        tmpfs_path=None,
        run_time_limit=None,
        env=None,
        persistent_volume_claims=None,
        tolerations=None,
        labels=None,
        shared_memory=None,
        port=None,
        name_pattern=None,
        qos=None,
        annotations=None,
        security_context=None,
    ):
        if env is None:
            env = {}
        job = (
            KubernetesClient()
            .job(
                generate_name=name_pattern,
                namespace=namespace,
                service_account=service_account,
                secrets=secrets,
                node_selector=node_selector,
                command=self._command(
                    flow_name=flow_name,
                    run_id=run_id,
                    step_name=step_name,
                    task_id=task_id,
                    attempt=attempt,
                    code_package_metadata=code_package_metadata,
                    code_package_url=code_package_url,
                    step_cmds=[step_cli],
                ),
                image=docker_image,
                image_pull_policy=docker_image_pull_policy,
                image_pull_secrets=image_pull_secrets,
                cpu=cpu,
                memory=memory,
                disk=disk,
                gpu=gpu,
                gpu_vendor=gpu_vendor,
                timeout_in_seconds=run_time_limit,
                # Retries are handled by Metaflow runtime
                retries=0,
                step_name=step_name,
                tolerations=tolerations,
                labels=labels,
                annotations=annotations,
                use_tmpfs=use_tmpfs,
                tmpfs_tempdir=tmpfs_tempdir,
                tmpfs_size=tmpfs_size,
                tmpfs_path=tmpfs_path,
                persistent_volume_claims=persistent_volume_claims,
                shared_memory=shared_memory,
                port=port,
                qos=qos,
                security_context=security_context,
            )
            .environment_variable("METAFLOW_CODE_METADATA", code_package_metadata)
            .environment_variable("METAFLOW_CODE_SHA", code_package_sha)
            .environment_variable("METAFLOW_CODE_URL", code_package_url)
            .environment_variable("METAFLOW_CODE_DS", code_package_ds)
            .environment_variable("METAFLOW_USER", user)
            .environment_variable("METAFLOW_SERVICE_URL", SERVICE_INTERNAL_URL)
            .environment_variable(
                "METAFLOW_SERVICE_HEADERS",
                json.dumps(SERVICE_HEADERS),
            )
            .environment_variable("METAFLOW_DATASTORE_SYSROOT_S3", DATASTORE_SYSROOT_S3)
            .environment_variable("METAFLOW_DATATOOLS_S3ROOT", DATATOOLS_S3ROOT)
            .environment_variable("METAFLOW_DEFAULT_DATASTORE", self._datastore.TYPE)
            .environment_variable("METAFLOW_DEFAULT_METADATA", DEFAULT_METADATA)
            .environment_variable("METAFLOW_KUBERNETES_WORKLOAD", 1)
            .environment_variable(
                "METAFLOW_KUBERNETES_FETCH_EC2_METADATA", KUBERNETES_FETCH_EC2_METADATA
            )
            .environment_variable("METAFLOW_RUNTIME_ENVIRONMENT", "kubernetes")
            .environment_variable(
                "METAFLOW_DEFAULT_SECRETS_BACKEND_TYPE", DEFAULT_SECRETS_BACKEND_TYPE
            )
            .environment_variable("METAFLOW_CARD_S3ROOT", CARD_S3ROOT)
            .environment_variable(
                "METAFLOW_DEFAULT_AWS_CLIENT_PROVIDER", DEFAULT_AWS_CLIENT_PROVIDER
            )
            .environment_variable(
                "METAFLOW_DEFAULT_GCP_CLIENT_PROVIDER", DEFAULT_GCP_CLIENT_PROVIDER
            )
            .environment_variable(
                "METAFLOW_AWS_SECRETS_MANAGER_DEFAULT_REGION",
                AWS_SECRETS_MANAGER_DEFAULT_REGION,
            )
            .environment_variable(
                "METAFLOW_GCP_SECRET_MANAGER_PREFIX", GCP_SECRET_MANAGER_PREFIX
            )
            .environment_variable(
                "METAFLOW_AZURE_KEY_VAULT_PREFIX", AZURE_KEY_VAULT_PREFIX
            )
            .environment_variable("METAFLOW_S3_ENDPOINT_URL", S3_ENDPOINT_URL)
            .environment_variable(
                "METAFLOW_AZURE_STORAGE_BLOB_SERVICE_ENDPOINT",
                AZURE_STORAGE_BLOB_SERVICE_ENDPOINT,
            )
            .environment_variable(
                "METAFLOW_DATASTORE_SYSROOT_AZURE", DATASTORE_SYSROOT_AZURE
            )
            .environment_variable("METAFLOW_CARD_AZUREROOT", CARD_AZUREROOT)
            .environment_variable("METAFLOW_DATASTORE_SYSROOT_GS", DATASTORE_SYSROOT_GS)
            .environment_variable("METAFLOW_CARD_GSROOT", CARD_GSROOT)
            # support Metaflow sandboxes
            .environment_variable(
                "METAFLOW_INIT_SCRIPT", KUBERNETES_SANDBOX_INIT_SCRIPT
            )
            .environment_variable(
                "METAFLOW_KUBERNETES_SANDBOX_INIT_SCRIPT",
                KUBERNETES_SANDBOX_INIT_SCRIPT,
            )
            .environment_variable(
                "METAFLOW_ARGO_WORKFLOWS_KUBERNETES_SECRETS",
                ARGO_WORKFLOWS_KUBERNETES_SECRETS,
            )
            .environment_variable(
                "METAFLOW_ARGO_WORKFLOWS_ENV_VARS_TO_SKIP",
                ARGO_WORKFLOWS_ENV_VARS_TO_SKIP,
            )
            .environment_variable("METAFLOW_OTEL_ENDPOINT", OTEL_ENDPOINT)
            # Skip setting METAFLOW_DATASTORE_SYSROOT_LOCAL because metadata sync
            # between the local user instance and the remote Kubernetes pod
            # assumes metadata is stored in DATASTORE_LOCAL_DIR on the Kubernetes
            # pod; this happens when METAFLOW_DATASTORE_SYSROOT_LOCAL is NOT set (
            # see get_datastore_root_from_config in datastore/local.py).
        )

        # Temporary passing of *some* environment variables. Do not rely on this
        # mechanism as it will be removed in the near future
        for k, v in config_values():
            if k.startswith("METAFLOW_CONDA_") or k.startswith("METAFLOW_DEBUG_"):
                job.environment_variable(k, v)

        if S3_SERVER_SIDE_ENCRYPTION is not None:
            job.environment_variable(
                "METAFLOW_S3_SERVER_SIDE_ENCRYPTION", S3_SERVER_SIDE_ENCRYPTION
            )

        # Set environment variables to support metaflow.integrations.ArgoEvent
        job.environment_variable(
            "METAFLOW_ARGO_EVENTS_WEBHOOK_URL", ARGO_EVENTS_INTERNAL_WEBHOOK_URL
        )
        job.environment_variable("METAFLOW_ARGO_EVENTS_EVENT", ARGO_EVENTS_EVENT)
        job.environment_variable(
            "METAFLOW_ARGO_EVENTS_EVENT_BUS", ARGO_EVENTS_EVENT_BUS
        )
        job.environment_variable(
            "METAFLOW_ARGO_EVENTS_EVENT_SOURCE", ARGO_EVENTS_EVENT_SOURCE
        )
        job.environment_variable(
            "METAFLOW_ARGO_EVENTS_SERVICE_ACCOUNT", ARGO_EVENTS_SERVICE_ACCOUNT
        )
        job.environment_variable(
            "METAFLOW_ARGO_EVENTS_WEBHOOK_AUTH",
            ARGO_EVENTS_WEBHOOK_AUTH,
        )

        tmpfs_enabled = use_tmpfs or (tmpfs_size and not use_tmpfs)
        if tmpfs_enabled and tmpfs_tempdir:
            job.environment_variable("METAFLOW_TEMPDIR", tmpfs_path)

        for name, value in env.items():
            job.environment_variable(name, value)
        # Add job specific labels
        system_labels = {
            "app.kubernetes.io/name": "metaflow-task",
            "app.kubernetes.io/part-of": "metaflow",
        }
        for name, value in system_labels.items():
            job.label(name, value)

        # Add job specific annotations not set in the decorator.
        system_annotations = {
            "metaflow/flow_name": flow_name,
            "metaflow/run_id": run_id,
            "metaflow/step_name": step_name,
            "metaflow/task_id": task_id,
            "metaflow/attempt": attempt,
            "metaflow/user": user,
        }
        if current.get("project_name"):
            system_annotations.update(
                {
                    "metaflow/project_name": current.project_name,
                    "metaflow/branch_name": current.branch_name,
                    "metaflow/project_flow_name": current.project_flow_name,
                }
            )

        for name, value in system_annotations.items():
            job.annotation(name, value)

        return job

    def create_k8sjob(self, job):
        return job.create()

    def wait(self, stdout_location, stderr_location, echo=None):
        def update_delay(secs_since_start):
            # this sigmoid function reaches
            # - 0.1 after 11 minutes
            # - 0.5 after 15 minutes
            # - 1.0 after 23 minutes
            # in other words, the user will see very frequent updates
            # during the first 10 minutes
            sigmoid = 1.0 / (1.0 + math.exp(-0.01 * secs_since_start + 9.0))
            return 0.5 + sigmoid * 30.0

        def wait_for_launch(job):
            status = job.status
            echo(
                "Task is starting (%s)..." % status,
                "stderr",
                job_id=job.id,
            )
            t = time.time()
            start_time = time.time()
            while job.is_waiting:
                new_status = job.status
                if status != new_status or (time.time() - t) > 30:
                    status = new_status
                    echo(
                        "Task is starting (%s)..." % status,
                        "stderr",
                        job_id=job.id,
                    )
                    t = time.time()
                time.sleep(update_delay(time.time() - start_time))

        prefix = lambda: b"[%s] " % util.to_bytes(self._job.id)

        stdout_tail = get_log_tailer(stdout_location, self._datastore.TYPE)
        stderr_tail = get_log_tailer(stderr_location, self._datastore.TYPE)

        # 1) Loop until the job has started
        wait_for_launch(self._job)

        # 2) Tail logs until the job has finished
        self._output_final_logs = False

        def _has_updates():
            if self._job.is_running:
                return True
            # Make sure to output final tail for a job that has finished.
            if not self._output_final_logs:
                self._output_final_logs = True
                return True
            return False

        tail_logs(
            prefix=prefix(),
            stdout_tail=stdout_tail,
            stderr_tail=stderr_tail,
            echo=echo,
            has_log_updates=_has_updates,
        )
        # 3) Fetch remaining logs
        #
        # It is possible that we exit the loop above before all logs have been
        # shown.
        #
        # TODO : If we notice Kubernetes failing to upload logs to S3,
        #        we can add a HEAD request here to ensure that the file
        #        exists prior to calling S3Tail and note the user about
        #        truncated logs if it doesn't.
        # TODO : For hard crashes, we can fetch logs from the pod.
        if self._job.has_failed:
            exit_code, reason = self._job.reason
            msg = next(
                msg
                for msg in [
                    reason,
                    "Task crashed",
                ]
                if msg is not None
            )
            if exit_code:
                if int(exit_code) == 139:
                    raise KubernetesException("Task failed with a segmentation fault.")
                if int(exit_code) == 137:
                    raise KubernetesException(
                        "Task ran out of memory. "
                        "Increase the available memory by specifying "
                        "@resource(memory=...) for the step. "
                    )
                if int(exit_code) == 134:
                    raise KubernetesException("%s (exit code %s)" % (msg, exit_code))
                else:
                    msg = "%s (exit code %s)" % (msg, exit_code)
            raise KubernetesException(
                "%s. This could be a transient error. Use @retry to retry." % msg
            )

        exit_code, _ = self._job.reason
        echo(
            "Task finished with exit code %s." % exit_code,
            "stderr",
            job_id=self._job.id,
        )



from typing import Dict, Optional, Union, TYPE_CHECKING

from metaflow.exception import MetaflowException
from metaflow.user_configs.config_parameters import (
    resolve_delayed_evaluator,
    unpack_delayed_evaluator,
)

from .common import ClassPath_Trie

if TYPE_CHECKING:
    import metaflow.flowspec
    import metaflow.user_decorators.mutable_flow


class FlowMutatorMeta(type):
    _all_registered_decorators = ClassPath_Trie()
    _do_not_register = set()
    _import_modules = set()

    def __new__(mcs, name, bases, namespace):
        cls = super().__new__(mcs, name, bases, namespace)
        cls.decorator_name = getattr(
            cls, "_decorator_name", f"{cls.__module__}.{cls.__name__}"
        )
        if not cls.__module__.startswith("metaflow.") and not cls.__module__.startswith(
            "metaflow_extensions."
        ):
            mcs._import_modules.add(cls.__module__)

        if name == "FlowMutator" or cls.decorator_name in mcs._do_not_register:
            return cls

        # We inject a __init_subclass__ method so we can figure out if there
        # are subclasses. We want to register as decorators only the ones that do
        # not have a subclass. The logic is that everything is registered and if
        # a subclass shows up, we will unregister the parent class leaving only those
        # classes that do not have any subclasses registered.
        @classmethod
        def do_unregister(cls_, **_kwargs):
            for base in cls_.__bases__:
                if isinstance(base, FlowMutatorMeta):
                    # If the base is a FlowMutatorMeta, we unregister it
                    # so that we don't have any decorators that are not the
                    # most derived one.
                    mcs._all_registered_decorators.remove(base.decorator_name)
                    # Also make sure we don't register again
                    mcs._do_not_register.add(base.decorator_name)

        cls.__init_subclass__ = do_unregister
        mcs._all_registered_decorators.insert(cls.decorator_name, cls)
        return cls

    @classmethod
    def all_decorators(mcs) -> Dict[str, "FlowMutatorMeta"]:
        mcs._check_init()
        return mcs._all_registered_decorators.get_unique_prefixes()

    def __str__(cls):
        return "FlowMutator(%s)" % cls.decorator_name

    @classmethod
    def get_decorator_by_name(
        mcs, decorator_name: str
    ) -> Optional[Union["FlowDecoratorMeta", "metaflow.decorators.Decorator"]]:
        """
        Get a decorator by its name.

        Parameters
        ----------
        decorator_name: str
            The name of the decorator to retrieve.

        Returns
        -------
        Optional[FlowDecoratorMeta]
            The decorator class if found, None otherwise.
        """
        mcs._check_init()
        return mcs._all_registered_decorators.unique_prefix_value(decorator_name)

    @classmethod
    def get_decorator_name(mcs, decorator_type: type) -> Optional[str]:
        """
        Get the minimally unique classpath name for a decorator type.

        Parameters
        ----------
        decorator_type: type
            The type of the decorator to retrieve the name for.

        Returns
        -------
        Optional[str]
            The minimally unique classpath name if found, None otherwise.
        """
        mcs._check_init()
        return mcs._all_registered_decorators.unique_prefix_for_type(decorator_type)

    @classmethod
    def _check_init(mcs):
        # Delay importing STEP_DECORATORS until we actually need it
        if not mcs._all_registered_decorators.inited:
            from metaflow.plugins import FLOW_DECORATORS

            mcs._all_registered_decorators.init([(t.name, t) for t in FLOW_DECORATORS])


class FlowMutator(metaclass=FlowMutatorMeta):
    """
    Derive from this class to implement a flow mutator.

    A flow mutator allows you to introspect a flow and its included steps. You can
    then add parameters, configurations and decorators to the flow as well as modify
    any of its steps.
    use values available through configurations to determine how to mutate the flow.

    There are two main methods provided:
      - pre_mutate: called as early as possible right after configuration values are read.
      - mutate: called right after all the command line is parsed but before any
        Metaflow decorators are applied.

    The `mutate` method does not allow you to modify the flow itself but you can still
    modify the steps.
    """

    def __init__(self, *args, **kwargs):
        from ..flowspec import FlowSpecMeta

        self._flow_cls = None
        # Flow mutators are always statically defined (no way of passing them
        # on the command line or adding them via configs)
        self.statically_defined = True
        self.inserted_by = None

        # The arguments are actually passed to the init function for this decorator
        # and used in _graph_info
        self._args = args
        self._kwargs = kwargs
        if args and isinstance(args[0], (FlowMutator, FlowSpecMeta)):
            # This means the decorator is bare like @MyDecorator
            # and the first argument is the FlowSpec or another decorator (they
            # can be stacked)

            # Now set the flow class we apply to
            if isinstance(args[0], FlowSpecMeta):
                self._set_flow_cls(args[0])
            else:
                self._set_flow_cls(args[0]._flow_cls)
            self._args = self._args[1:]  # Remove the first argument

    def __mro_entries__(self, bases):
        # This is called in the following case:
        # @MyMutator
        # class MyBaseFlowSpec(FlowSpec):
        #   pass
        #
        # class MyFlow(MyBaseFlowSpec):
        #   pass
        #
        # MyBaseFlowSpec will be an object of type MyMutator which is not
        # great when inheriting from it. With this method, we ensure that the
        # base class will actually be MyBaseFlowSpec
        return (self._flow_cls,)

    def __call__(
        self, flow_spec: Optional["metaflow.flowspec.FlowSpecMeta"] = None
    ) -> "metaflow.flowspec.FlowSpecMeta":
        if flow_spec:
            if isinstance(flow_spec, FlowMutator):
                flow_spec = flow_spec._flow_cls
            return self._set_flow_cls(flow_spec)
        elif not self._flow_cls:
            # This means that somehow the initialization did not happen properly
            # so this may have been applied to a non flow
            raise MetaflowException("A FlowMutator can only be applied to a FlowSpec")
        # NOTA: This returns self._flow_cls() because the object in the case of
        # @FlowDecorator
        # class MyFlow(FlowSpec):
        #     pass
        # the object is a FlowDecorator and when the main function calls it, we end up
        # here and need to actually call the FlowSpec. This is not the case when using
        # a decorator with arguments because in the line above, we will have returned a
        # FlowSpec object. Previous solution was to use __get__ but this does not seem
        # to work properly.
        return self._flow_cls()

    def _set_flow_cls(
        self, flow_spec: "metaflow.flowspec.FlowSpecMeta"
    ) -> "metaflow.flowspec.FlowSpecMeta":
        from ..flowspec import _FlowState

        flow_spec._flow_state.setdefault(_FlowState.FLOW_MUTATORS, []).append(self)
        self._flow_cls = flow_spec
        return flow_spec

    def __str__(self):
        return str(self.__class__)

    def init(self, *args, **kwargs):
        """
        Implement this method if you wish for your FlowMutator to take in arguments.

        Your flow-mutator can then look like:

        @MyMutator(arg1, arg2)
        class MyFlow(FlowSpec):
            pass

        It is an error to use your mutator with arguments but not implement this method.

        """
        pass

    def external_init(self):
        # You can use config values in the arguments to a FlowMutator
        # so we resolve those as well
        self._args = [resolve_delayed_evaluator(arg) for arg in self._args]
        self._kwargs, _ = unpack_delayed_evaluator(self._kwargs)
        self._kwargs = {
            k: resolve_delayed_evaluator(v) for k, v in self._kwargs.items()
        }
        if self._args or self._kwargs:
            if "init" not in self.__class__.__dict__:
                raise MetaflowException(
                    "%s is used with arguments but does not implement init" % self
                )
        if "init" in self.__class__.__dict__:
            self.init(*self._args, **self._kwargs)

    def pre_mutate(
        self, mutable_flow: "metaflow.user_decorators.mutable_flow.MutableFlow"
    ) -> None:
        """
        Method called right after all configuration values are read.

        Parameters
        ----------
        mutable_flow : metaflow.user_decorators.mutable_flow.MutableFlow
            A representation of this flow
        """
        return None

    def mutate(
        self, mutable_flow: "metaflow.user_decorators.mutable_flow.MutableFlow"
    ) -> None:
        """
        Method called right before the first Metaflow step decorator is applied. This
        means that the command line, including all `--with` options has been parsed.

        Given how late this function is called, there are a few restrictions on what
        you can do; the following methods on MutableFlow are not allowed and calling
        them will result in an error:
          - add_parameter/remove_parameter
          - add_decorator/remove_decorator

        To call these methods, use the `pre_mutate` method instead.

        Parameters
        ----------
        mutable_flow : metaflow.user_decorators.mutable_flow.MutableFlow
            A representation of this flow
        """
        return None



import os
import json
import sys


def export_parameters(output_file):
    input = json.loads(os.environ.get("METAFLOW_PARAMETERS", "{}"))
    with open(output_file, "w") as f:
        for k in input:
            # Replace `-` with `_` is parameter names since `-` isn't an
            # allowed character for environment variables. cli.py will
            # correctly translate the replaced `-`s.
            f.write(
                "export METAFLOW_INIT_%s=%s\n"
                % (k.upper().replace("-", "_"), json.dumps(input[k]))
            )
    os.chmod(output_file, 509)


if __name__ == "__main__":
    export_parameters(sys.argv[1])



import json
import time
from threading import Thread

import requests

from metaflow.exception import MetaflowException
from metaflow.metaflow_config import SERVICE_HEADERS
from metaflow.sidecar import Message, MessageTypes

HB_URL_KEY = "hb_url"


class HeartBeatException(MetaflowException):
    headline = "Metaflow heart beat error"

    def __init__(self, msg):
        super(HeartBeatException, self).__init__(msg)


class MetadataHeartBeat(object):
    def __init__(self):
        self.headers = SERVICE_HEADERS
        self.req_thread = Thread(target=self._ping)
        self.req_thread.daemon = True
        self.default_frequency_secs = 10
        self.hb_url = None

    def process_message(self, msg):
        # type: (Message) -> None
        if msg.msg_type == MessageTypes.SHUTDOWN:
            self._shutdown()
        if not self.req_thread.is_alive():
            # set post url
            self.hb_url = msg.payload[HB_URL_KEY]
            # start thread
            self.req_thread.start()

    @classmethod
    def get_worker(cls):
        return cls

    def _ping(self):
        retry_counter = 0
        while True:
            try:
                frequency_secs = self._heartbeat()

                if frequency_secs is None or frequency_secs <= 0:
                    frequency_secs = self.default_frequency_secs

                time.sleep(frequency_secs)
                retry_counter = 0
            except HeartBeatException as e:
                print(e)
                retry_counter = retry_counter + 1
                time.sleep(1.5**retry_counter)

    def _heartbeat(self):
        if self.hb_url is not None:
            try:
                response = requests.post(
                    url=self.hb_url, data="{}", headers=self.headers.copy()
                )
            except requests.exceptions.ConnectionError as e:
                raise HeartBeatException(
                    "HeartBeat request (%s) failed" " (ConnectionError)" % (self.hb_url)
                )
            except requests.exceptions.Timeout as e:
                raise HeartBeatException(
                    "HeartBeat request (%s) failed" " (Timeout)" % (self.hb_url)
                )
            except requests.exceptions.RequestException as e:
                raise HeartBeatException(
                    "HeartBeat request (%s) failed"
                    " (RequestException) %s" % (self.hb_url, str(e))
                )
            # Unfortunately, response.json() returns a string that we need
            # to cast to json; however when the request encounters an error
            # the return type is a json blob :/
            if response.status_code == 200:
                return json.loads(response.json()).get("wait_time_in_seconds")
            else:
                raise HeartBeatException(
                    "HeartBeat request (%s) failed"
                    " (code %s): %s"
                    % (self.hb_url, response.status_code, response.text)
                )
        return None

    def _shutdown(self):
        # attempts sending one last heartbeat
        self._heartbeat()



"""
Local backend

Execute the flow with a native runtime
using local / remote processes
"""

from __future__ import print_function
import json
import os
import sys
import fcntl
import re
import tempfile
import time
import subprocess
from datetime import datetime
from enum import Enum
from io import BytesIO
from itertools import chain
from functools import partial
from concurrent import futures

from typing import Dict, Tuple
from metaflow.datastore.exceptions import DataException
from contextlib import contextmanager

from . import get_namespace
from .metadata_provider import MetaDatum
from .metaflow_config import FEAT_ALWAYS_UPLOAD_CODE_PACKAGE, MAX_ATTEMPTS, UI_URL
from .exception import (
    MetaflowException,
    MetaflowInternalError,
    METAFLOW_EXIT_DISALLOW_RETRY,
)
from . import procpoll
from .datastore import TaskDataStoreSet
from .debug import debug
from .decorators import flow_decorators
from .flowspec import _FlowState
from .mflog import mflog, RUNTIME_LOG_SOURCE
from .util import to_unicode, compress_list, unicode_type
from .clone_util import clone_task_helper
from .unbounded_foreach import (
    CONTROL_TASK_TAG,
    UBF_CONTROL,
    UBF_TASK,
)

from .user_configs.config_options import ConfigInput
from .user_configs.config_parameters import dump_config_values

import metaflow.tracing as tracing

MAX_WORKERS = 16
MAX_NUM_SPLITS = 100
MAX_LOG_SIZE = 1024 * 1024
POLL_TIMEOUT = 1000  # ms
PROGRESS_INTERVAL = 300  # s
# The following is a list of the (data) artifacts used by the runtime while
# executing a flow. These are prefetched during the resume operation by
# leveraging the TaskDataStoreSet.
PREFETCH_DATA_ARTIFACTS = [
    "_foreach_stack",
    "_iteration_stack",
    "_task_ok",
    "_transition",
    "_control_mapper_tasks",
    "_control_task_is_mapper_zero",
]
RESUME_POLL_SECONDS = 60


class LoopBehavior(Enum):
    NONE = "none"
    ENTERING = "entering"
    EXITING = "exiting"
    LOOPING = "looping"


# Runtime must use logsource=RUNTIME_LOG_SOURCE for all loglines that it
# formats according to mflog. See a comment in mflog.__init__
mflog_msg = partial(mflog.decorate, RUNTIME_LOG_SOURCE)

# TODO option: output dot graph periodically about execution


class NativeRuntime(object):
    def __init__(
        self,
        flow,
        graph,
        flow_datastore,
        metadata,
        environment,
        package,
        logger,
        entrypoint,
        event_logger,
        monitor,
        run_id=None,
        clone_run_id=None,
        clone_only=False,
        reentrant=False,
        steps_to_rerun=None,
        max_workers=MAX_WORKERS,
        max_num_splits=MAX_NUM_SPLITS,
        max_log_size=MAX_LOG_SIZE,
        resume_identifier=None,
        skip_decorator_hooks=False,
    ):
        if run_id is None:
            self._run_id = metadata.new_run_id()
        else:
            self._run_id = run_id
            metadata.register_run_id(run_id)

        self._flow = flow
        self._graph = graph
        self._flow_datastore = flow_datastore
        self._metadata = metadata
        self._environment = environment
        self._package = package
        self._logger = logger
        self._max_workers = max_workers
        self._active_tasks = dict()  # Key: step name;
        # value: [number of running tasks, number of done tasks]
        # Special key 0 is total number of running tasks
        self._active_tasks[0] = 0
        self._unprocessed_steps = set([n.name for n in self._graph])
        self._max_num_splits = max_num_splits
        self._max_log_size = max_log_size
        self._params_task = None
        self._entrypoint = entrypoint
        self.event_logger = event_logger
        self._monitor = monitor
        self._resume_identifier = resume_identifier

        self._clone_run_id = clone_run_id
        self._clone_only = clone_only
        self._cloned_tasks = []
        self._ran_or_scheduled_task_index = set()
        self._reentrant = reentrant
        self._run_url = None
        self._skip_decorator_hooks = skip_decorator_hooks

        # If steps_to_rerun is specified, we will not clone them in resume mode.
        self._steps_to_rerun = steps_to_rerun or {}
        # sorted_nodes are in topological order already, so we only need to
        # iterate through the nodes once to get a stable set of rerun steps.
        for step_name in self._graph.sorted_nodes:
            if step_name in self._steps_to_rerun:
                out_funcs = self._graph[step_name].out_funcs or []
                for next_step in out_funcs:
                    self._steps_to_rerun.add(next_step)

        self._origin_ds_set = None
        if clone_run_id:
            # resume logic
            # 0. If clone_run_id is specified, attempt to clone all the
            # successful tasks from the flow with `clone_run_id`. And run the
            # unsuccessful or not-run steps in the regular fashion.
            # 1. With _find_origin_task, for every task in the current run, we
            # find the equivalent task in `clone_run_id` using
            # pathspec_index=run/step:[index] and verify if this task can be
            # cloned.
            # 2. If yes, we fire off a clone-only task which copies the
            # metadata from the `clone_origin` to pathspec=run/step/task to
            # mimmick that the resumed run looks like an actual run.
            # 3. All steps that couldn't be cloned (either unsuccessful or not
            # run) are run as regular tasks.
            # Lastly, to improve the performance of the cloning process, we
            # leverage the TaskDataStoreSet abstraction to prefetch the
            # entire DAG of `clone_run_id` and relevant data artifacts
            # (see PREFETCH_DATA_ARTIFACTS) so that the entire runtime can
            # access the relevant data from cache (instead of going to the datastore
            # after the first prefetch).
            logger(
                "Gathering required information to resume run (this may take a bit of time)..."
            )
            self._origin_ds_set = TaskDataStoreSet(
                flow_datastore,
                clone_run_id,
                prefetch_data_artifacts=PREFETCH_DATA_ARTIFACTS,
            )
        self._run_queue = []
        self._poll = procpoll.make_poll()
        self._workers = {}  # fd -> subprocess mapping
        self._finished = {}
        self._is_cloned = {}
        # NOTE: In case of unbounded foreach, we need the following to schedule
        # the (sibling) mapper tasks  of the control task (in case of resume);
        # and ensure that the join tasks runs only if all dependent tasks have
        # finished.
        self._control_num_splits = {}  # control_task -> num_splits mapping

        if not self._skip_decorator_hooks:
            for step in flow:
                for deco in step.decorators:
                    deco.runtime_init(flow, graph, package, self._run_id)

    def _new_task(self, step, input_paths=None, **kwargs):
        if input_paths is None:
            may_clone = True
        else:
            may_clone = all(self._is_cloned[path] for path in input_paths)

        if step in self._steps_to_rerun:
            may_clone = False

        if step == "_parameters" or self._skip_decorator_hooks:
            decos = []
        else:
            decos = getattr(self._flow, step).decorators

        return Task(
            self._flow_datastore,
            self._flow,
            step,
            self._run_id,
            self._metadata,
            self._environment,
            self._entrypoint,
            self.event_logger,
            self._monitor,
            input_paths=input_paths,
            may_clone=may_clone,
            clone_run_id=self._clone_run_id,
            clone_only=self._clone_only,
            reentrant=self._reentrant,
            origin_ds_set=self._origin_ds_set,
            decos=decos,
            logger=self._logger,
            resume_identifier=self._resume_identifier,
            **kwargs,
        )

    @property
    def run_id(self):
        return self._run_id

    def persist_constants(self, task_id=None):
        self._params_task = self._new_task("_parameters", task_id=task_id)
        if not self._params_task.is_cloned:
            self._params_task.persist(self._flow)

        self._is_cloned[self._params_task.path] = self._params_task.is_cloned

    def should_skip_clone_only_execution(self):
        (
            should_skip_clone_only_execution,
            skip_reason,
        ) = self._should_skip_clone_only_execution()
        if should_skip_clone_only_execution:
            self._logger(skip_reason, system_msg=True)
            return True
        return False

    @contextmanager
    def run_heartbeat(self):
        self._metadata.start_run_heartbeat(self._flow.name, self._run_id)
        yield
        self._metadata.stop_heartbeat()

    def print_workflow_info(self):
        self._run_url = (
            "%s/%s/%s" % (UI_URL.rstrip("/"), self._flow.name, self._run_id)
            if UI_URL
            else None
        )

        if self._run_url:
            self._logger(
                "Workflow starting (run-id %s), see it in the UI at %s"
                % (
                    self._run_id,
                    self._run_url,
                ),
                system_msg=True,
            )
        else:
            self._logger(
                "Workflow starting (run-id %s):" % self._run_id, system_msg=True
            )

    def _should_skip_clone_only_execution(self):
        if self._clone_only and self._params_task:
            if self._params_task.resume_done():
                return True, "Resume already complete. Skip clone-only execution."
            if not self._params_task.is_resume_leader():
                return (
                    True,
                    "Not resume leader under resume execution. Skip clone-only execution.",
                )
        return False, None

    def clone_task(
        self,
        step_name,
        task_id,
        pathspec_index,
        cloned_task_pathspec_index,
        finished_tuple,
        iteration_tuple,
        ubf_context,
        generate_task_obj,
        verbose=False,
    ):
        try:
            new_task_id = task_id
            if generate_task_obj:
                task = self._new_task(step_name, pathspec_index=pathspec_index)
                if ubf_context:
                    task.ubf_context = ubf_context
                new_task_id = task.task_id
                self._cloned_tasks.append(task)
                self._ran_or_scheduled_task_index.add(cloned_task_pathspec_index)
                task_pathspec = "{}/{}/{}".format(self._run_id, step_name, new_task_id)
            else:
                task_pathspec = "{}/{}/{}".format(self._run_id, step_name, new_task_id)
                Task.clone_pathspec_mapping[task_pathspec] = "{}/{}/{}".format(
                    self._clone_run_id, step_name, task_id
                )
            if verbose:
                self._logger(
                    "Cloning task from {}/{}/{}/{} to {}/{}/{}/{}".format(
                        self._flow.name,
                        self._clone_run_id,
                        step_name,
                        task_id,
                        self._flow.name,
                        self._run_id,
                        step_name,
                        new_task_id,
                    ),
                    system_msg=True,
                )
            clone_task_helper(
                self._flow.name,
                self._clone_run_id,
                self._run_id,
                step_name,
                task_id,  # origin_task_id
                new_task_id,
                self._flow_datastore,
                self._metadata,
                origin_ds_set=self._origin_ds_set,
            )
            self._finished[(step_name, finished_tuple, iteration_tuple)] = task_pathspec
            self._is_cloned[task_pathspec] = True
        except Exception as e:
            self._logger(
                "Cloning {}/{}/{}/{} failed with error: {}".format(
                    self._flow.name, self._clone_run_id, step_name, task_id, str(e)
                )
            )

    def clone_original_run(self, generate_task_obj=False, verbose=True):
        self._logger(
            "Cloning {}/{}".format(self._flow.name, self._clone_run_id),
            system_msg=True,
        )

        inputs = []

        ubf_mapper_tasks_to_clone = set()
        ubf_control_tasks = set()
        # We only clone ubf mapper tasks if the control task is complete.
        # Here we need to check which control tasks are complete, and then get the corresponding
        # mapper tasks.
        for task_ds in self._origin_ds_set:
            _, step_name, task_id = task_ds.pathspec.split("/")
            pathspec_index = task_ds.pathspec_index
            if task_ds["_task_ok"] and step_name != "_parameters":
                # Control task contains "_control_mapper_tasks" but, in the case of
                # @parallel decorator, the control task is also a mapper task so we
                # need to distinguish this using _control_task_is_mapper_zero
                control_mapper_tasks = (
                    []
                    if "_control_mapper_tasks" not in task_ds
                    else task_ds["_control_mapper_tasks"]
                )
                if control_mapper_tasks:
                    if task_ds.get("_control_task_is_mapper_zero", False):
                        # Strip out the control task of list of mapper tasks
                        ubf_control_tasks.add(control_mapper_tasks[0])
                        ubf_mapper_tasks_to_clone.update(control_mapper_tasks[1:])
                    else:
                        ubf_mapper_tasks_to_clone.update(control_mapper_tasks)
                        # Since we only add mapper tasks here, if we are not in the list
                        # we are a control task
                        if task_ds.pathspec not in ubf_mapper_tasks_to_clone:
                            ubf_control_tasks.add(task_ds.pathspec)

        for task_ds in self._origin_ds_set:
            _, step_name, task_id = task_ds.pathspec.split("/")
            pathspec_index = task_ds.pathspec_index

            if (
                task_ds["_task_ok"]
                and step_name != "_parameters"
                and (step_name not in self._steps_to_rerun)
            ):
                # "_unbounded_foreach" is a special flag to indicate that the transition
                # is an unbounded foreach.
                # Both parent and splitted children tasks will have this flag set.
                # The splitted control/mapper tasks
                # are not foreach types because UBF is always followed by a join step.
                is_ubf_task = (
                    "_unbounded_foreach" in task_ds and task_ds["_unbounded_foreach"]
                ) and (self._graph[step_name].type != "foreach")

                is_ubf_control_task = task_ds.pathspec in ubf_control_tasks

                is_ubf_mapper_task = is_ubf_task and (not is_ubf_control_task)

                if is_ubf_mapper_task and (
                    task_ds.pathspec not in ubf_mapper_tasks_to_clone
                ):
                    # Skip copying UBF mapper tasks if control task is incomplete.
                    continue

                ubf_context = None
                if is_ubf_task:
                    ubf_context = "ubf_test" if is_ubf_mapper_task else "ubf_control"

                finished_tuple = tuple(
                    [s._replace(value=0) for s in task_ds.get("_foreach_stack", ())]
                )
                iteration_tuple = tuple(task_ds.get("_iteration_stack", ()))
                cloned_task_pathspec_index = pathspec_index.split("/")[1]
                if task_ds.get("_control_task_is_mapper_zero", False):
                    # Replace None with index 0 for control task as it is part of the
                    # UBF (as a mapper as well)
                    finished_tuple = finished_tuple[:-1] + (
                        finished_tuple[-1]._replace(index=0),
                    )
                    # We need this reverse override though because when we check
                    # if a task has been cloned in _queue_push, the index will be None
                    # because the _control_task_is_mapper_zero is set in the control
                    # task *itself* and *not* in the one that is launching the UBF nest.
                    # This means that _translate_index will use None.
                    cloned_task_pathspec_index = re.sub(
                        r"(\[(?:\d+, ?)*)0\]",
                        lambda m: (m.group(1) or "[") + "None]",
                        cloned_task_pathspec_index,
                    )

                inputs.append(
                    (
                        step_name,
                        task_id,
                        pathspec_index,
                        cloned_task_pathspec_index,
                        finished_tuple,
                        iteration_tuple,
                        is_ubf_mapper_task,
                        ubf_context,
                    )
                )

        with futures.ThreadPoolExecutor(max_workers=self._max_workers) as executor:
            all_tasks = [
                executor.submit(
                    self.clone_task,
                    step_name,
                    task_id,
                    pathspec_index,
                    cloned_task_pathspec_index,
                    finished_tuple,
                    iteration_tuple,
                    ubf_context=ubf_context,
                    generate_task_obj=generate_task_obj and (not is_ubf_mapper_task),
                    verbose=verbose,
                )
                for (
                    step_name,
                    task_id,
                    pathspec_index,
                    cloned_task_pathspec_index,
                    finished_tuple,
                    iteration_tuple,
                    is_ubf_mapper_task,
                    ubf_context,
                ) in inputs
            ]
            _, _ = futures.wait(all_tasks)
        self._logger(
            "{}/{} cloned!".format(self._flow.name, self._clone_run_id), system_msg=True
        )
        self._params_task.mark_resume_done()

    def execute(self):
        if len(self._cloned_tasks) > 0:
            # mutable list storing the cloned tasks.
            self._run_queue = []
            self._active_tasks[0] = 0
        else:
            if self._params_task:
                self._queue_push("start", {"input_paths": [self._params_task.path]})
            else:
                self._queue_push("start", {})

        progress_tstamp = time.time()
        with tempfile.NamedTemporaryFile(mode="w", encoding="utf-8") as config_file:
            # Configurations are passed through a file to avoid overloading the
            # command-line. We only need to create this file once and it can be reused
            # for any task launch
            config_value = dump_config_values(self._flow)
            if config_value:
                json.dump(config_value, config_file)
                config_file.flush()
                self._config_file_name = config_file.name
            else:
                self._config_file_name = None
            try:
                # main scheduling loop
                exception = None
                while (
                    self._run_queue or self._active_tasks[0] > 0 or self._cloned_tasks
                ):
                    # 1. are any of the current workers finished?
                    if self._cloned_tasks:
                        finished_tasks = []

                        # For loops (right now just recursive steps), we need to find
                        # the exact frontier because if we queue all "successors" to all
                        # the finished iterations, we would incorrectly launch multiple
                        # successors. We therefore have to strip out all non-last
                        # iterations *per* foreach branch.
                        idx_per_finished_id = (
                            {}
                        )  # type: Dict[Tuple[str, Tuple[int, ...], Tuple[int, Tuple[int, ...]]]]
                        for task in self._cloned_tasks:
                            step_name, foreach_stack, iteration_stack = task.finished_id
                            existing_task_idx = idx_per_finished_id.get(
                                (step_name, foreach_stack), None
                            )
                            if existing_task_idx is not None:
                                len_diff = len(iteration_stack) - len(
                                    existing_task_idx[1]
                                )
                                # In this case, we need to keep only the latest iteration
                                if (
                                    len_diff == 0
                                    and iteration_stack > existing_task_idx[1]
                                ) or len_diff == -1:
                                    # We remove the one we currently have and replace
                                    # by this one. The second option means that we are
                                    # adding the finished iteration marker.
                                    existing_task = finished_tasks[existing_task_idx[0]]
                                    # These are the first two lines of _queue_tasks
                                    # We still consider the tasks finished so we need
                                    # to update state to be clean.
                                    self._finished[existing_task.finished_id] = (
                                        existing_task.path
                                    )
                                    self._is_cloned[existing_task.path] = (
                                        existing_task.is_cloned
                                    )

                                    finished_tasks[existing_task_idx[0]] = task
                                    idx_per_finished_id[(step_name, foreach_stack)] = (
                                        existing_task_idx[0],
                                        iteration_stack,
                                    )
                                elif (
                                    len_diff == 0
                                    and iteration_stack < existing_task_idx[1]
                                ) or len_diff == 1:
                                    # The second option is when we have already marked
                                    # the end of the iteration in self._finished and
                                    # are now seeing a previous iteration.
                                    # We just mark the task as finished but we don't
                                    # put it in the finished_tasks list to pass to
                                    # the _queue_tasks function
                                    self._finished[task.finished_id] = task.path
                                    self._is_cloned[task.path] = task.is_cloned
                                else:
                                    raise MetaflowInternalError(
                                        "Unexpected recursive cloned tasks -- "
                                        "this is a bug, please report it."
                                    )
                            else:
                                # New entry
                                finished_tasks.append(task)
                                idx_per_finished_id[(step_name, foreach_stack)] = (
                                    len(finished_tasks) - 1,
                                    iteration_stack,
                                )

                        # reset the list of cloned tasks and let poll_workers handle
                        # the remaining transition
                        self._cloned_tasks = []
                    else:
                        finished_tasks = list(self._poll_workers())
                    # 2. push new tasks triggered by the finished tasks to the queue
                    self._queue_tasks(finished_tasks)
                    # 3. if there are available worker slots, pop and start tasks
                    #    from the queue.
                    self._launch_workers()

                    if time.time() - progress_tstamp > PROGRESS_INTERVAL:
                        progress_tstamp = time.time()
                        tasks_print = ", ".join(
                            [
                                "%s (%d running; %d done)" % (k, v[0], v[1])
                                for k, v in self._active_tasks.items()
                                if k != 0 and v[0] > 0
                            ]
                        )
                        if self._active_tasks[0] == 0:
                            msg = "No tasks are running."
                        else:
                            if self._active_tasks[0] == 1:
                                msg = "1 task is running: "
                            else:
                                msg = "%d tasks are running: " % self._active_tasks[0]
                            msg += "%s." % tasks_print

                        self._logger(msg, system_msg=True)

                        if len(self._run_queue) == 0:
                            msg = "No tasks are waiting in the queue."
                        else:
                            if len(self._run_queue) == 1:
                                msg = "1 task is waiting in the queue: "
                            else:
                                msg = "%d tasks are waiting in the queue." % len(
                                    self._run_queue
                                )

                        self._logger(msg, system_msg=True)
                        if len(self._unprocessed_steps) > 0:
                            if len(self._unprocessed_steps) == 1:
                                msg = "%s step has not started" % (
                                    next(iter(self._unprocessed_steps)),
                                )
                            else:
                                msg = "%d steps have not started: " % len(
                                    self._unprocessed_steps
                                )
                                msg += "%s." % ", ".join(self._unprocessed_steps)
                            self._logger(msg, system_msg=True)

            except KeyboardInterrupt as ex:
                self._logger("Workflow interrupted.", system_msg=True, bad=True)
                self._killall()
                exception = ex
                raise
            except Exception as ex:
                self._logger("Workflow failed.", system_msg=True, bad=True)
                self._killall()
                exception = ex
                raise
            finally:
                # on finish clean tasks
                if not self._skip_decorator_hooks:
                    for step in self._flow:
                        for deco in step.decorators:
                            deco.runtime_finished(exception)
                self._run_exit_hooks()

        # assert that end was executed and it was successful
        if ("end", (), ()) in self._finished:
            if self._run_url:
                self._logger(
                    "Done! See the run in the UI at %s" % self._run_url,
                    system_msg=True,
                )
            else:
                self._logger("Done!", system_msg=True)
        elif self._clone_only:
            self._logger(
                "Clone-only resume complete -- only previously successful steps were "
                "cloned; no new tasks executed!",
                system_msg=True,
            )
            self._params_task.mark_resume_done()
        else:
            raise MetaflowInternalError(
                "The *end* step was not successful by the end of flow."
            )

    def _run_exit_hooks(self):
        try:
            exit_hook_decos = self._flow._flow_decorators.get("exit_hook", [])
            if not exit_hook_decos:
                return

            successful = ("end", (), ()) in self._finished or self._clone_only
            pathspec = f"{self._graph.name}/{self._run_id}"
            flow_file = self._environment.get_environment_info()["script"]

            def _call(fn_name):
                try:
                    result = (
                        subprocess.check_output(
                            args=[
                                sys.executable,
                                "-m",
                                "metaflow.plugins.exit_hook.exit_hook_script",
                                flow_file,
                                fn_name,
                                pathspec,
                            ],
                            env=os.environ,
                        )
                        .decode()
                        .strip()
                    )
                    print(result)
                except subprocess.CalledProcessError as e:
                    print(f"[exit_hook] Hook '{fn_name}' failed with error: {e}")
                except Exception as e:
                    print(f"[exit_hook] Unexpected error in hook '{fn_name}': {e}")

            # Call all exit hook functions regardless of individual failures
            for fn_name in [
                name
                for deco in exit_hook_decos
                for name in (deco.success_hooks if successful else deco.error_hooks)
            ]:
                _call(fn_name)

        except Exception as ex:
            pass  # do not fail due to exit hooks for whatever reason.

    def _killall(self):
        # If we are here, all children have received a signal and are shutting down.
        # We want to give them an opportunity to do so and then kill
        live_workers = set(self._workers.values())
        now = int(time.time())
        self._logger(
            "Terminating %d active tasks..." % len(live_workers),
            system_msg=True,
            bad=True,
        )
        while live_workers and int(time.time()) - now < 5:
            # While not all workers are dead and we have waited less than 5 seconds
            live_workers = [worker for worker in live_workers if not worker.clean()]
        if live_workers:
            self._logger(
                "Killing %d remaining tasks after having waited for %d seconds -- "
                "some tasks may not exit cleanly"
                % (len(live_workers), int(time.time()) - now),
                system_msg=True,
                bad=True,
            )
            for worker in live_workers:
                worker.kill()
        self._logger("Flushing logs...", system_msg=True, bad=True)
        # give killed workers a chance to flush their logs to datastore
        for _ in range(3):
            list(self._poll_workers())

    # Given the current task information (task_index), the type of transition,
    # and the split index, return the new task index.
    def _translate_index(
        self, task, next_step, type, split_index=None, loop_mode=LoopBehavior.NONE
    ):
        match = re.match(r"^(.+)\[(.*)\]\[(.*)\]$", task.task_index)
        old_match = re.match(r"^(.+)\[(.*)\]$", task.task_index)
        if match:
            _, foreach_index, iteration_index = match.groups()
            # Convert foreach_index to a list of integers
            if len(foreach_index) > 0:
                foreach_index = foreach_index.split(",")
            else:
                foreach_index = []
            # Ditto for iteration_index
            if len(iteration_index) > 0:
                iteration_index = iteration_index.split(",")
            else:
                iteration_index = []
        elif old_match:
            _, foreach_index = old_match.groups()
            # Convert foreach_index to a list of integers
            if len(foreach_index) > 0:
                foreach_index = foreach_index.split(",")
            else:
                foreach_index = []
            # Legacy case fallback. No iteration index exists for these runs.
            iteration_index = []
        else:
            raise ValueError(
                "Index not in the format of {run_id}/{step_name}[{foreach_index}][{iteration_index}]"
            )
        if loop_mode == LoopBehavior.NONE:
            # Check if we are entering a looping construct. Right now, only recursive
            # steps are looping constructs
            next_step_node = self._graph[next_step]
            if (
                next_step_node.type == "split-switch"
                and next_step in next_step_node.out_funcs
            ):
                loop_mode = LoopBehavior.ENTERING

        # Update iteration_index
        if loop_mode == LoopBehavior.ENTERING:
            # We are entering a loop, so we add a new iteration level
            iteration_index.append("0")
        elif loop_mode == LoopBehavior.EXITING:
            iteration_index = iteration_index[:-1]
        elif loop_mode == LoopBehavior.LOOPING:
            if len(iteration_index) == 0:
                raise MetaflowInternalError(
                    "In looping mode but there is no iteration index"
                )
            iteration_index[-1] = str(int(iteration_index[-1]) + 1)
        iteration_index = ",".join(iteration_index)

        if type == "linear":
            return "%s[%s][%s]" % (next_step, ",".join(foreach_index), iteration_index)
        elif type == "join":
            indices = []
            if len(foreach_index) > 0:
                indices = foreach_index[:-1]
            return "%s[%s][%s]" % (next_step, ",".join(indices), iteration_index)
        elif type == "split":
            foreach_index.append(str(split_index))
            return "%s[%s][%s]" % (next_step, ",".join(foreach_index), iteration_index)

    # Store the parameters needed for task creation, so that pushing on items
    # onto the run_queue is an inexpensive operation.
    def _queue_push(self, step, task_kwargs, index=None):
        # In the case of cloning, we set all the cloned tasks as the
        # finished tasks when pushing tasks using _queue_tasks. This means that we
        # could potentially try to push the same task multiple times (for example
        # if multiple parents of a join are cloned). We therefore keep track of what
        # has executed (been cloned) or what has been scheduled and avoid scheduling
        # it again.
        if index:
            if index in self._ran_or_scheduled_task_index:
                # It has already run or been scheduled
                return
            # Note that we are scheduling this to run
            self._ran_or_scheduled_task_index.add(index)
        self._run_queue.insert(0, (step, task_kwargs))
        # For foreaches, this will happen multiple time but is ok, becomes a no-op
        self._unprocessed_steps.discard(step)

    def _queue_pop(self):
        return self._run_queue.pop() if self._run_queue else (None, {})

    def _queue_task_join(self, task, next_steps):
        # if the next step is a join, we need to check that
        # all input tasks for the join have finished before queuing it.

        # CHECK: this condition should be enforced by the linter but
        # let's assert that the assumption holds
        if len(next_steps) > 1:
            msg = (
                "Step *{step}* transitions to a join and another "
                "step. The join must be the only transition."
            )
            raise MetaflowInternalError(task, msg.format(step=task.step))
        else:
            next_step = next_steps[0]

        unbounded_foreach = not task.results.is_none("_unbounded_foreach")

        if unbounded_foreach:
            # Before we queue the join, do some post-processing of runtime state
            # (_finished, _is_cloned) for the (sibling) mapper tasks.
            # Update state of (sibling) mapper tasks for control task.
            if task.ubf_context == UBF_CONTROL:
                mapper_tasks = task.results.get("_control_mapper_tasks")
                if not mapper_tasks:
                    msg = (
                        "Step *{step}* has a control task which didn't "
                        "specify the artifact *_control_mapper_tasks* for "
                        "the subsequent *{join}* step."
                    )
                    raise MetaflowInternalError(
                        msg.format(step=task.step, join=next_steps[0])
                    )
                elif not (
                    isinstance(mapper_tasks, list)
                    and isinstance(mapper_tasks[0], unicode_type)
                ):
                    msg = (
                        "Step *{step}* has a control task which didn't "
                        "specify the artifact *_control_mapper_tasks* as a "
                        "list of strings but instead specified it as {typ} "
                        "with elements of {elem_typ}."
                    )
                    raise MetaflowInternalError(
                        msg.format(
                            step=task.step,
                            typ=type(mapper_tasks),
                            elem_type=type(mapper_tasks[0]),
                        )
                    )
                num_splits = len(mapper_tasks)
                self._control_num_splits[task.path] = num_splits

                # If the control task is cloned, all mapper tasks should have been cloned
                # as well, so we no longer need to handle cloning of mapper tasks in runtime.

                # Update _finished if we are not cloned. If we were cloned, we already
                # updated _finished with the new tasks. Note that the *value* of mapper
                # tasks is incorrect and contains the pathspec of the *cloned* run
                # but we don't use it for anything. We could look to clean it up though
                if not task.is_cloned:
                    _, foreach_stack, iteration_stack = task.finished_id
                    top = foreach_stack[-1]
                    bottom = list(foreach_stack[:-1])
                    for i in range(num_splits):
                        s = tuple(bottom + [top._replace(index=i)])
                        self._finished[(task.step, s, iteration_stack)] = mapper_tasks[
                            i
                        ]
                        self._is_cloned[mapper_tasks[i]] = False

            # Find and check status of control task and retrieve its pathspec
            # for retrieving unbounded foreach cardinality.
            _, foreach_stack, iteration_stack = task.finished_id
            top = foreach_stack[-1]
            bottom = list(foreach_stack[:-1])
            s = tuple(bottom + [top._replace(index=None)])

            # UBF control can also be the first task of the list. Then
            # it will have index=0 instead of index=None.
            if task.results.get("_control_task_is_mapper_zero", False):
                s = tuple(bottom + [top._replace(index=0)])
            control_path = self._finished.get((task.step, s, iteration_stack))
            if control_path:
                # Control task was successful.
                # Additionally check the state of (sibling) mapper tasks as well
                # (for the sake of resume) before queueing join task.
                num_splits = self._control_num_splits[control_path]
                required_tasks = []
                for i in range(num_splits):
                    s = tuple(bottom + [top._replace(index=i)])
                    required_tasks.append(
                        self._finished.get((task.step, s, iteration_stack))
                    )

                if all(required_tasks):
                    index = self._translate_index(task, next_step, "join")
                    # all tasks to be joined are ready. Schedule the next join step.
                    self._queue_push(
                        next_step,
                        {"input_paths": required_tasks, "join_type": "foreach"},
                        index,
                    )
        else:
            # matching_split is the split-parent of the finished task
            matching_split = self._graph[self._graph[next_step].split_parents[-1]]
            _, foreach_stack, iteration_stack = task.finished_id

            direct_parents = set(self._graph[next_step].in_funcs)

            # next step is a foreach join
            if matching_split.type == "foreach":

                def siblings(foreach_stack):
                    top = foreach_stack[-1]
                    bottom = list(foreach_stack[:-1])
                    for index in range(top.num_splits):
                        yield tuple(bottom + [top._replace(index=index)])

                # required tasks are all split-siblings of the finished task
                required_tasks = list(
                    filter(
                        lambda x: x is not None,
                        [
                            self._finished.get((p, s, iteration_stack))
                            for p in direct_parents
                            for s in siblings(foreach_stack)
                        ],
                    )
                )
                required_count = task.finished_id[1][-1].num_splits
                join_type = "foreach"
                index = self._translate_index(task, next_step, "join")
            else:
                # next step is a split
                required_tasks = list(
                    filter(
                        lambda x: x is not None,
                        [
                            self._finished.get((p, foreach_stack, iteration_stack))
                            for p in direct_parents
                        ],
                    )
                )

                required_count = len(matching_split.out_funcs)
                join_type = "linear"
                index = self._translate_index(task, next_step, "linear")
            if len(required_tasks) == required_count:
                # We have all the required previous tasks to schedule a join
                self._queue_push(
                    next_step,
                    {"input_paths": required_tasks, "join_type": join_type},
                    index,
                )

    def _queue_task_switch(self, task, next_steps, is_recursive):
        chosen_step = next_steps[0]

        loop_mode = LoopBehavior.NONE
        if is_recursive:
            if chosen_step != task.step:
                # We are exiting a loop
                loop_mode = LoopBehavior.EXITING
            else:
                # We are staying in the loop
                loop_mode = LoopBehavior.LOOPING
        index = self._translate_index(task, chosen_step, "linear", None, loop_mode)
        self._queue_push(chosen_step, {"input_paths": [task.path]}, index)

    def _queue_task_foreach(self, task, next_steps):
        # CHECK: this condition should be enforced by the linter but
        # let's assert that the assumption holds
        if len(next_steps) > 1:
            msg = (
                "Step *{step}* makes a foreach split but it defines "
                "multiple transitions. Specify only one transition "
                "for foreach."
            )
            raise MetaflowInternalError(msg.format(step=task.step))
        else:
            next_step = next_steps[0]

        unbounded_foreach = not task.results.is_none("_unbounded_foreach")
        if unbounded_foreach:
            # Need to push control process related task.
            ubf_iter_name = task.results.get("_foreach_var")
            ubf_iter = task.results.get(ubf_iter_name)
            # UBF control task has no split index, hence "None" as place holder.

            if task.results.get("_control_task_is_mapper_zero", False):
                index = self._translate_index(task, next_step, "split", 0)
            else:
                index = self._translate_index(task, next_step, "split", None)
            self._queue_push(
                next_step,
                {
                    "input_paths": [task.path],
                    "ubf_context": UBF_CONTROL,
                    "ubf_iter": ubf_iter,
                },
                index,
            )
        else:
            num_splits = task.results["_foreach_num_splits"]
            if num_splits > self._max_num_splits:
                msg = (
                    "Foreach in step *{step}* yielded {num} child steps "
                    "which is more than the current maximum of {max} "
                    "children. You can raise the maximum with the "
                    "--max-num-splits option. "
                )
                raise TaskFailed(
                    task,
                    msg.format(
                        step=task.step, num=num_splits, max=self._max_num_splits
                    ),
                )

            # schedule all splits
            for i in range(num_splits):
                index = self._translate_index(task, next_step, "split", i)
                self._queue_push(
                    next_step,
                    {"split_index": str(i), "input_paths": [task.path]},
                    index,
                )

    def _queue_tasks(self, finished_tasks):
        # finished tasks include only successful tasks
        for task in finished_tasks:
            self._finished[task.finished_id] = task.path
            self._is_cloned[task.path] = task.is_cloned

            # CHECK: ensure that runtime transitions match with
            # statically inferred transitions. Make an exception for control
            # tasks, where we just rely on static analysis since we don't
            # execute user code.
            trans = task.results.get("_transition")
            if trans:
                next_steps = trans[0]
                foreach = trans[1]
            else:
                next_steps = []
                foreach = None
            expected = self._graph[task.step].out_funcs

            if self._graph[task.step].type == "split-switch":
                is_recursive = task.step in self._graph[task.step].out_funcs
                if len(next_steps) != 1:
                    msg = (
                        "Switch step *{step}* should transition to exactly "
                        "one step at runtime, but got: {actual}"
                    )
                    raise MetaflowInternalError(
                        msg.format(step=task.step, actual=", ".join(next_steps))
                    )
                if next_steps[0] not in expected:
                    msg = (
                        "Switch step *{step}* transitioned to unexpected "
                        "step *{actual}*. Expected one of: {expected}"
                    )
                    raise MetaflowInternalError(
                        msg.format(
                            step=task.step,
                            actual=next_steps[0],
                            expected=", ".join(expected),
                        )
                    )
                # When exiting a recursive loop, we mark that the loop itself has
                # finished by adding a special entry in self._finished which has
                # an iteration stack that is shorter (ie: we are out of the loop) so
                # that we can then find it when looking at successor tasks to launch.
                if is_recursive and next_steps[0] != task.step:
                    step_name, finished_tuple, iteration_tuple = task.finished_id
                    self._finished[
                        (step_name, finished_tuple, iteration_tuple[:-1])
                    ] = task.path
            elif next_steps != expected:
                msg = (
                    "Based on static analysis of the code, step *{step}* "
                    "was expected to transition to step(s) *{expected}*. "
                    "However, when the code was executed, self.next() was "
                    "called with *{actual}*. Make sure there is only one "
                    "unconditional self.next() call in the end of your "
                    "step. "
                )
                raise MetaflowInternalError(
                    msg.format(
                        step=task.step,
                        expected=", ".join(expected),
                        actual=", ".join(next_steps),
                    )
                )

            # Different transition types require different treatment
            if any(self._graph[f].type == "join" for f in next_steps):
                # Next step is a join
                self._queue_task_join(task, next_steps)
            elif foreach:
                # Next step is a foreach child
                self._queue_task_foreach(task, next_steps)
            elif self._graph[task.step].type == "split-switch":
                # Current step is switch - queue the chosen step
                self._queue_task_switch(task, next_steps, is_recursive)
            else:
                # Next steps are normal linear steps
                for step in next_steps:
                    index = self._translate_index(task, step, "linear")
                    self._queue_push(step, {"input_paths": [task.path]}, index)

    def _poll_workers(self):
        if self._workers:
            for event in self._poll.poll(POLL_TIMEOUT):
                worker = self._workers.get(event.fd)
                if worker:
                    if event.can_read:
                        worker.read_logline(event.fd)
                    if event.is_terminated:
                        returncode = worker.terminate()

                        for fd in worker.fds():
                            self._poll.remove(fd)
                            del self._workers[fd]
                        step_counts = self._active_tasks[worker.task.step]
                        step_counts[0] -= 1  # One less task for this step is running
                        step_counts[1] += 1  # ... and one more completed.
                        # We never remove from self._active_tasks because it is possible
                        # for all currently running task for a step to complete but
                        # for others to still be queued up.
                        self._active_tasks[0] -= 1

                        task = worker.task
                        if returncode:
                            # worker did not finish successfully
                            if (
                                worker.cleaned
                                or returncode == METAFLOW_EXIT_DISALLOW_RETRY
                            ):
                                self._logger(
                                    "This failed task will not be retried.",
                                    system_msg=True,
                                )
                            else:
                                if (
                                    task.retries
                                    < task.user_code_retries + task.error_retries
                                ):
                                    self._retry_worker(worker)
                                else:
                                    raise TaskFailed(task)
                        else:
                            # worker finished successfully
                            yield task

    def _launch_workers(self):
        while self._run_queue and self._active_tasks[0] < self._max_workers:
            step, task_kwargs = self._queue_pop()
            # Initialize the task (which can be expensive using remote datastores)
            # before launching the worker so that cost is amortized over time, instead
            # of doing it during _queue_push.
            if (
                FEAT_ALWAYS_UPLOAD_CODE_PACKAGE
                and "METAFLOW_CODE_SHA" not in os.environ
            ):
                # We check if the code package is uploaded and, if so, we set the
                # environment variables that will cause the metadata service to
                # register the code package with the task created in _new_task below
                code_sha = self._package.package_sha(timeout=0.01)
                if code_sha:
                    os.environ["METAFLOW_CODE_SHA"] = code_sha
                    os.environ["METAFLOW_CODE_URL"] = self._package.package_url()
                    os.environ["METAFLOW_CODE_DS"] = self._flow_datastore.TYPE
                    os.environ["METAFLOW_CODE_METADATA"] = (
                        self._package.package_metadata
                    )

            task = self._new_task(step, **task_kwargs)
            self._launch_worker(task)

    def _retry_worker(self, worker):
        worker.task.retries += 1
        if worker.task.retries >= MAX_ATTEMPTS:
            # any results with an attempt ID >= MAX_ATTEMPTS will be ignored
            # by datastore, so running a task with such a retry_could would
            # be pointless and dangerous
            raise MetaflowInternalError(
                "Too many task attempts (%d)! "
                "MAX_ATTEMPTS exceeded." % worker.task.retries
            )

        worker.task.new_attempt()
        self._launch_worker(worker.task)

    def _launch_worker(self, task):
        if self._clone_only and not task.is_cloned:
            # We don't launch a worker here
            self._logger(
                "Not executing non-cloned task for step '%s' in clone-only resume"
                % "/".join([task.flow_name, task.run_id, task.step]),
                system_msg=True,
            )
            return

        worker = Worker(task, self._max_log_size, self._config_file_name)
        for fd in worker.fds():
            self._workers[fd] = worker
            self._poll.add(fd)
        active_step_counts = self._active_tasks.setdefault(task.step, [0, 0])

        # We have an additional task for this step running
        active_step_counts[0] += 1

        # One more task actively running
        self._active_tasks[0] += 1


class Task(object):
    clone_pathspec_mapping = {}

    def __init__(
        self,
        flow_datastore,
        flow,
        step,
        run_id,
        metadata,
        environment,
        entrypoint,
        event_logger,
        monitor,
        input_paths=None,
        may_clone=False,
        clone_run_id=None,
        clone_only=False,
        reentrant=False,
        origin_ds_set=None,
        decos=None,
        logger=None,
        # Anything below this is passed as part of kwargs
        split_index=None,
        ubf_context=None,
        ubf_iter=None,
        join_type=None,
        task_id=None,
        resume_identifier=None,
        pathspec_index=None,
    ):
        self.step = step
        self.flow = flow
        self.flow_name = flow.name
        self.run_id = run_id
        self.task_id = None
        self._path = None
        self.input_paths = input_paths
        self.split_index = split_index
        self.ubf_context = ubf_context
        self.ubf_iter = ubf_iter
        self.decos = [] if decos is None else decos
        self.entrypoint = entrypoint
        self.environment = environment
        self.environment_type = self.environment.TYPE
        self.clone_run_id = clone_run_id
        self.clone_origin = None
        self.origin_ds_set = origin_ds_set
        self.metadata = metadata
        self.event_logger = event_logger
        self.monitor = monitor

        self._logger = logger

        self.retries = 0
        self.user_code_retries = 0
        self.error_retries = 0

        self.tags = metadata.sticky_tags
        self.event_logger_type = self.event_logger.TYPE
        self.monitor_type = monitor.TYPE

        self.metadata_type = metadata.TYPE
        self.datastore_type = flow_datastore.TYPE
        self._flow_datastore = flow_datastore
        self.datastore_sysroot = flow_datastore.datastore_root
        self._results_ds = None

        # Only used in clone-only resume.
        self._is_resume_leader = None
        self._resume_done = None
        self._resume_identifier = resume_identifier
        origin = None
        if clone_run_id and may_clone:
            origin = self._find_origin_task(clone_run_id, join_type, pathspec_index)
        if origin and origin["_task_ok"]:
            # At this point, we know we are going to clone
            self._is_cloned = True

            task_id_exists_already = False
            task_completed = False
            if reentrant:
                # A re-entrant clone basically allows multiple concurrent processes
                # to perform the clone at the same time to the same new run id. Let's
                # assume two processes A and B both simultaneously calling
                # `resume --reentrant --run-id XX`.
                # We want to guarantee that:
                #   - All incomplete tasks are cloned exactly once.
                # To achieve this, we will select a resume leader and let it clone the
                # entire execution graph. This ensures that we only write once to the
                # datastore and metadata.
                #
                # We use the cloned _parameter task's task-id as the "key" to synchronize
                # on. We try to "register" this new task-id (or rather the full pathspec
                # <run>/<step>/<taskid>) with the metadata service which will indicate
                # if we actually registered it or if it existed already. If we did manage
                # to register it (_parameter task), we are the "elected resume leader"
                # in essence and proceed to clone. If we didn't, we just wait to make
                # sure the entire clone execution is fully done (ie: the clone is finished).
                if task_id is not None:
                    # Sanity check -- this should never happen. We cannot allow
                    # for explicit task-ids because in the reentrant case, we use the
                    # cloned task's id as the new task's id.
                    raise MetaflowInternalError(
                        "Reentrant clone-only resume does not allow for explicit task-id"
                    )

                if resume_identifier:
                    self.log(
                        "Resume identifier is %s." % resume_identifier,
                        system_msg=True,
                    )
                else:
                    raise MetaflowInternalError(
                        "Reentrant clone-only resume needs a resume identifier."
                    )
                # We will use the same task_id as the original task
                # to use it effectively as a synchronization key
                clone_task_id = origin.task_id
                # Make sure the task-id is a non-integer to not clash with task ids
                # assigned by the metadata provider. If this is an integer, we
                # add some string to it. It doesn't matter what we add as long as it is
                # consistent.
                try:
                    clone_task_int = int(clone_task_id)
                    clone_task_id = "resume-%d" % clone_task_int
                except ValueError:
                    pass

                # If _get_task_id returns True it means the task already existed, so
                # we wait for it.
                task_id_exists_already = self._get_task_id(clone_task_id)

                # We may not have access to task datastore on first resume attempt, but
                # on later resume attempt, we should check if the resume task is complete
                # or not. This is to fix the issue where the resume leader was killed
                # unexpectedly during cloning and never mark task complete.
                try:
                    task_completed = self.results["_task_ok"]
                except DataException as e:
                    pass
            else:
                self._get_task_id(task_id)

            # Store the mapping from current_pathspec -> origin_pathspec which
            # will be useful for looking up origin_ds_set in find_origin_task.
            self.clone_pathspec_mapping[self._path] = origin.pathspec
            if self.step == "_parameters":
                # In the _parameters task, we need to resolve who is the resume leader.
                self._is_resume_leader = False
                resume_leader = None

                if task_id_exists_already:
                    # If the task id already exists, we need to check if current task is the resume leader in previous attempt.
                    ds = self._flow_datastore.get_task_datastore(
                        self.run_id, self.step, self.task_id
                    )
                    if not ds["_task_ok"]:
                        raise MetaflowInternalError(
                            "Externally cloned _parameters task did not succeed"
                        )

                    # Check if we should be the resume leader (maybe from previous attempt).
                    # To avoid the edge case where the resume leader is selected but has not
                    # yet written the _resume_leader metadata, we will wait for a few seconds.
                    # We will wait for resume leader for at most 3 times.
                    for _ in range(3):
                        if ds.has_metadata("_resume_leader", add_attempt=False):
                            resume_leader = ds.load_metadata(
                                ["_resume_leader"], add_attempt=False
                            )["_resume_leader"]
                            self._is_resume_leader = resume_leader == resume_identifier
                        else:
                            self.log(
                                "Waiting for resume leader to be selected. Sleeping ...",
                                system_msg=True,
                            )
                            time.sleep(3)
                else:
                    # If the task id does not exist, current task is the resume leader.
                    resume_leader = resume_identifier
                    self._is_resume_leader = True

                if reentrant:
                    if resume_leader:
                        self.log(
                            "Resume leader is %s." % resume_leader,
                            system_msg=True,
                        )
                    else:
                        raise MetaflowInternalError(
                            "Can not determine the resume leader in distributed resume mode."
                        )

                if self._is_resume_leader:
                    if reentrant:
                        self.log(
                            "Selected as the reentrant clone leader.",
                            system_msg=True,
                        )
                    # Clone in place without relying on run_queue.
                    self.new_attempt()
                    self._ds.clone(origin)
                    # Set the resume leader be the task that calls the resume (first task to clone _parameters task).
                    # We will always set resume leader regardless whether we are in distributed resume case or not.
                    if resume_identifier:
                        self._ds.save_metadata(
                            {"_resume_leader": resume_identifier}, add_attempt=False
                        )

                    self._ds.done()
                else:
                    # Wait for the resume leader to complete
                    while True:
                        ds = self._flow_datastore.get_task_datastore(
                            self.run_id, self.step, self.task_id
                        )

                        # Check if resume is complete. Resume leader will write the done file.
                        self._resume_done = ds.has_metadata(
                            "_resume_done", add_attempt=False
                        )

                        if self._resume_done:
                            break

                        self.log(
                            "Waiting for resume leader to complete. Sleeping for %ds..."
                            % RESUME_POLL_SECONDS,
                            system_msg=True,
                        )
                        time.sleep(RESUME_POLL_SECONDS)
                    self.log(
                        "_parameters clone completed by resume leader", system_msg=True
                    )
            else:
                # Only leader can reach non-parameter steps in resume.

                # Store the origin pathspec in clone_origin so this can be run
                # as a task by the runtime.
                self.clone_origin = origin.pathspec
                # Save a call to creating the results_ds since its same as origin.
                self._results_ds = origin

                # If the task is already completed in new run, we don't need to clone it.
                self._should_skip_cloning = task_completed
                if self._should_skip_cloning:
                    self.log(
                        "Skipping cloning of previously run task %s"
                        % self.clone_origin,
                        system_msg=True,
                    )
                else:
                    self.log(
                        "Cloning previously run task %s" % self.clone_origin,
                        system_msg=True,
                    )
        else:
            self._is_cloned = False
            if clone_only:
                # We are done -- we don't proceed to create new task-ids
                return
            self._get_task_id(task_id)

        # Open the output datastore only if the task is not being cloned.
        if not self._is_cloned:
            self.new_attempt()
            for deco in decos:
                deco.runtime_task_created(
                    self._ds,
                    task_id,
                    split_index,
                    input_paths,
                    self._is_cloned,
                    ubf_context,
                )

                # determine the number of retries of this task
                user_code_retries, error_retries = deco.step_task_retry_count()
                if user_code_retries is None and error_retries is None:
                    # This signals the runtime that the task doesn't want any
                    # retries indifferent to other decorator opinions.
                    # NOTE: This is needed since we don't statically disallow
                    # specifying `@retry` in combination with decorators which
                    # implement `unbounded_foreach` semantics. This allows for
                    # ergonomic user invocation of `--with retry`; instead
                    # choosing to specially handle this way in the runtime.
                    self.user_code_retries = None
                    self.error_retries = None
                if (
                    self.user_code_retries is not None
                    and self.error_retries is not None
                ):
                    self.user_code_retries = max(
                        self.user_code_retries, user_code_retries
                    )
                    self.error_retries = max(self.error_retries, error_retries)
            if self.user_code_retries is None and self.error_retries is None:
                self.user_code_retries = 0
                self.error_retries = 0

    def new_attempt(self):
        self._ds = self._flow_datastore.get_task_datastore(
            self.run_id, self.step, self.task_id, attempt=self.retries, mode="w"
        )
        self._ds.init_task()

    def log(self, msg, system_msg=False, pid=None, timestamp=True):
        if pid:
            prefix = "[%s (pid %s)] " % (self._path, pid)
        else:
            prefix = "[%s] " % self._path

        self._logger(msg, head=prefix, system_msg=system_msg, timestamp=timestamp)
        sys.stdout.flush()

    def is_resume_leader(self):
        assert (
            self.step == "_parameters"
        ), "Only _parameters step can check resume leader."
        return self._is_resume_leader

    def resume_done(self):
        assert (
            self.step == "_parameters"
        ), "Only _parameters step can check wheather resume is complete."
        return self._resume_done

    def mark_resume_done(self):
        assert (
            self.step == "_parameters"
        ), "Only _parameters step can mark resume as done."
        assert self.is_resume_leader(), "Only resume leader can mark resume as done."

        # Mark the resume as done. This is called at the end of the resume flow and after
        # the _parameters step was successfully cloned, so we need to 'dangerously' save
        # this done file, but the risk should be minimal.
        self._ds._dangerous_save_metadata_post_done(
            {"_resume_done": True}, add_attempt=False
        )

    def _get_task_id(self, task_id):
        already_existed = True
        tags = []
        if self.ubf_context == UBF_CONTROL:
            tags = [CONTROL_TASK_TAG]
        # Register Metaflow tasks.
        if task_id is None:
            task_id = str(
                self.metadata.new_task_id(self.run_id, self.step, sys_tags=tags)
            )
            already_existed = False
        else:
            # task_id is preset only by persist_constants().
            already_existed = not self.metadata.register_task_id(
                self.run_id,
                self.step,
                task_id,
                0,
                sys_tags=tags,
            )

        self.task_id = task_id
        self._path = "%s/%s/%s" % (self.run_id, self.step, self.task_id)
        return already_existed

    def _find_origin_task(self, clone_run_id, join_type, pathspec_index=None):
        if pathspec_index:
            origin = self.origin_ds_set.get_with_pathspec_index(pathspec_index)
            return origin
        elif self.step == "_parameters":
            pathspec = "%s/_parameters[]" % clone_run_id
            origin = self.origin_ds_set.get_with_pathspec_index(pathspec)

            if origin is None:
                # This is just for usability: We could rerun the whole flow
                # if an unknown clone_run_id is provided but probably this is
                # not what the user intended, so raise a warning
                raise MetaflowException(
                    "Resume could not find run id *%s*" % clone_run_id
                )
            else:
                return origin
        else:
            # all inputs must have the same foreach stack, so we can safely
            # pick the first one
            parent_pathspec = self.input_paths[0]
            origin_parent_pathspec = self.clone_pathspec_mapping[parent_pathspec]
            parent = self.origin_ds_set.get_with_pathspec(origin_parent_pathspec)
            # Parent should be non-None since only clone the child if the parent
            # was successfully cloned.
            foreach_stack = parent["_foreach_stack"]
            if join_type == "foreach":
                # foreach-join pops the topmost index
                index = ",".join(str(s.index) for s in foreach_stack[:-1])
            elif self.split_index or self.ubf_context == UBF_CONTROL:
                # foreach-split pushes a new index
                index = ",".join(
                    [str(s.index) for s in foreach_stack] + [str(self.split_index)]
                )
            else:
                # all other transitions keep the parent's foreach stack intact
                index = ",".join(str(s.index) for s in foreach_stack)
            pathspec = "%s/%s[%s]" % (clone_run_id, self.step, index)
            return self.origin_ds_set.get_with_pathspec_index(pathspec)

    @property
    def path(self):
        return self._path

    @property
    def results(self):
        if self._results_ds:
            return self._results_ds
        else:
            self._results_ds = self._flow_datastore.get_task_datastore(
                self.run_id, self.step, self.task_id
            )
            return self._results_ds

    @property
    def task_index(self):
        _, task_index = self.results.pathspec_index.split("/")
        return task_index

    @property
    def finished_id(self):
        # note: id is not available before the task has finished.
        # Index already identifies the task within the foreach and loop.
        # We will remove foreach value so that it is easier to
        # identify siblings within a foreach.
        foreach_stack_tuple = tuple(
            [s._replace(value=0) for s in self.results["_foreach_stack"]]
        )
        # _iteration_stack requires a fallback, as it does not exist for runs before v2.17.4
        iteration_stack_tuple = tuple(self.results.get("_iteration_stack", []))
        return (self.step, foreach_stack_tuple, iteration_stack_tuple)

    @property
    def is_cloned(self):
        return self._is_cloned

    @property
    def should_skip_cloning(self):
        return self._should_skip_cloning

    def persist(self, flow):
        # this is used to persist parameters before the start step
        flow._task_ok = flow._success = True
        flow._foreach_stack = []
        self._ds.persist(flow)
        self._ds.done()

    def save_logs(self, logtype_to_logs):
        self._ds.save_logs(RUNTIME_LOG_SOURCE, logtype_to_logs)

    def save_metadata(self, name, metadata):
        self._ds.save_metadata({name: metadata})

    def __str__(self):
        return " ".join(self._args)


class TaskFailed(MetaflowException):
    headline = "Step failure"

    def __init__(self, task, msg=""):
        body = "Step *%s* (task-id %s) failed" % (task.step, task.task_id)
        if msg:
            body = "%s: %s" % (body, msg)
        else:
            body += "."

        super(TaskFailed, self).__init__(body)


class TruncatedBuffer(object):
    def __init__(self, name, maxsize):
        self.name = name
        self._maxsize = maxsize
        self._buffer = BytesIO()
        self._size = 0
        self._eof = False

    def write(self, bytedata, system_msg=False):
        if system_msg:
            self._buffer.write(bytedata)
        elif not self._eof:
            if self._size + len(bytedata) < self._maxsize:
                self._buffer.write(bytedata)
                self._size += len(bytedata)
            else:
                msg = b"[TRUNCATED - MAXIMUM LOG FILE SIZE REACHED]\n"
                self._buffer.write(mflog_msg(msg))
                self._eof = True

    def get_bytes(self):
        return self._buffer.getvalue()

    def get_buffer(self):
        self._buffer.seek(0)
        return self._buffer


class CLIArgs(object):
    """
    Container to allow decorators modify the command line parameters
    for step execution in StepDecorator.runtime_step_cli().
    """

    def __init__(self, task):
        self.task = task
        self.entrypoint = list(task.entrypoint)
        step_obj = getattr(self.task.flow, self.task.step)
        self.top_level_options = {
            "quiet": True,
            "metadata": self.task.metadata_type,
            "environment": self.task.environment_type,
            "datastore": self.task.datastore_type,
            "pylint": False,
            "event-logger": self.task.event_logger_type,
            "monitor": self.task.monitor_type,
            "datastore-root": self.task.datastore_sysroot,
            "with": [
                deco.make_decorator_spec()
                for deco in chain(
                    self.task.decos,
                    step_obj.wrappers,
                    step_obj.config_decorators,
                )
                if not deco.statically_defined and deco.inserted_by is None
            ],
        }

        # FlowDecorators can define their own top-level options. They are
        # responsible for adding their own top-level options and values through
        # the get_top_level_options() hook.
        for deco in flow_decorators(self.task.flow):
            self.top_level_options.update(deco.get_top_level_options())

        # We also pass configuration options using the kv.<name> syntax which will cause
        # the configuration options to be loaded from the CONFIG file (or local-config-file
        # in the case of the local runtime)
        configs = self.task.flow._flow_state.get(_FlowState.CONFIGS)
        if configs:
            self.top_level_options["config-value"] = [
                (k, ConfigInput.make_key_name(k)) for k in configs
            ]

        self.commands = ["step"]
        self.command_args = [self.task.step]
        self.command_options = {
            "run-id": task.run_id,
            "task-id": task.task_id,
            "input-paths": compress_list(task.input_paths),
            "split-index": task.split_index,
            "retry-count": task.retries,
            "max-user-code-retries": task.user_code_retries,
            "tag": task.tags,
            "namespace": get_namespace() or "",
            "ubf-context": task.ubf_context,
        }
        self.env = {}

    def get_args(self):
        # TODO: Make one with dict_to_cli_options; see cli_args.py for more detail
        def _options(mapping):
            for k, v in mapping.items():
                # None or False arguments are ignored
                # v needs to be explicitly False, not falsy, e.g. 0 is an acceptable value
                if v is None or v is False:
                    continue

                # we need special handling for 'with' since it is a reserved
                # keyword in Python, so we call it 'decospecs' in click args
                if k == "decospecs":
                    k = "with"
                k = k.replace("_", "-")
                v = v if isinstance(v, (list, tuple, set)) else [v]
                for value in v:
                    yield "--%s" % k
                    if not isinstance(value, bool):
                        value = value if isinstance(value, tuple) else (value,)
                        for vv in value:
                            yield to_unicode(vv)

        args = list(self.entrypoint)
        args.extend(_options(self.top_level_options))
        args.extend(self.commands)
        args.extend(self.command_args)

        args.extend(_options(self.command_options))
        return args

    def get_env(self):
        return self.env

    def __str__(self):
        return " ".join(self.get_args())


class Worker(object):
    def __init__(self, task, max_logs_size, config_file_name):
        self.task = task
        self._config_file_name = config_file_name
        self._proc = self._launch()

        if task.retries > task.user_code_retries:
            self.task.log(
                "Task fallback is starting to handle the failure.",
                system_msg=True,
                pid=self._proc.pid,
            )
        elif not task.is_cloned:
            suffix = " (retry)." if task.retries else "."
            self.task.log(
                "Task is starting" + suffix, system_msg=True, pid=self._proc.pid
            )

        self._stdout = TruncatedBuffer("stdout", max_logs_size)
        self._stderr = TruncatedBuffer("stderr", max_logs_size)

        self._logs = {
            self._proc.stderr.fileno(): (self._proc.stderr, self._stderr),
            self._proc.stdout.fileno(): (self._proc.stdout, self._stdout),
        }

        self._encoding = sys.stdout.encoding or "UTF-8"
        self.killed = False  # Killed indicates that the task was forcibly killed
        # with SIGKILL by the master process.
        # A killed task is always considered cleaned
        self.cleaned = False  # A cleaned task is one that is shutting down and has been
        # noticed by the runtime and queried for its state (whether or
        # not it is properly shut down)

    def _launch(self):
        args = CLIArgs(self.task)
        env = dict(os.environ)

        if self.task.clone_run_id:
            args.command_options["clone-run-id"] = self.task.clone_run_id

        if self.task.is_cloned and self.task.clone_origin:
            args.command_options["clone-only"] = self.task.clone_origin
            # disabling sidecars for cloned tasks due to perf reasons
            args.top_level_options["event-logger"] = "nullSidecarLogger"
            args.top_level_options["monitor"] = "nullSidecarMonitor"
        else:
            # decorators may modify the CLIArgs object in-place
            for deco in self.task.decos:
                deco.runtime_step_cli(
                    args,
                    self.task.retries,
                    self.task.user_code_retries,
                    self.task.ubf_context,
                )

        # Add user configurations using a file to avoid using up too much space on the
        # command line
        if self._config_file_name:
            args.top_level_options["local-config-file"] = self._config_file_name
        # Pass configuration options
        env.update(args.get_env())
        env["PYTHONUNBUFFERED"] = "x"
        tracing.inject_tracing_vars(env)
        # NOTE bufsize=1 below enables line buffering which is required
        # by read_logline() below that relies on readline() not blocking
        # print('running', args)
        cmdline = args.get_args()
        debug.subcommand_exec(cmdline)
        return subprocess.Popen(
            cmdline,
            env=env,
            bufsize=1,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )

    def emit_log(self, msg, buf, system_msg=False):
        if mflog.is_structured(msg):
            res = mflog.parse(msg)
            if res:
                # parsing successful
                plain = res.msg
                timestamp = res.utc_tstamp
                if res.should_persist:
                    # in special circumstances we may receive structured
                    # loglines that haven't been properly persisted upstream.
                    # This is the case if, for example, a task crashes in an external
                    # system and we retrieve the remaining logs after the crash.
                    # Those lines are marked with a special tag, should_persist,
                    # which we process here
                    buf.write(mflog.unset_should_persist(msg))
            else:
                # parsing failed, corrupted logline. Print it as-is
                timestamp = datetime.utcnow()
                plain = msg
        else:
            # If the line isn't formatted with mflog already, we format it here.
            plain = msg
            timestamp = datetime.utcnow()
            # store unformatted loglines in the buffer that will be
            # persisted, assuming that all previously formatted loglines have
            # been already persisted at the source.
            buf.write(mflog_msg(msg, now=timestamp), system_msg=system_msg)
        text = plain.strip().decode(self._encoding, errors="replace")
        self.task.log(
            text,
            pid=self._proc.pid,
            timestamp=mflog.utc_to_local(timestamp),
            system_msg=system_msg,
        )

    def read_logline(self, fd):
        fileobj, buf = self._logs[fd]
        # readline() below should never block thanks to polling and
        # line buffering. If it does, things will deadlock
        line = fileobj.readline()
        if line:
            self.emit_log(line, buf)
            return True
        else:
            return False

    def fds(self):
        return (self._proc.stderr.fileno(), self._proc.stdout.fileno())

    def clean(self):
        if self.killed:
            return True
        if not self.cleaned:
            for fileobj, buf in self._logs.values():
                msg = b"[KILLED BY ORCHESTRATOR]\n"
                self.emit_log(msg, buf, system_msg=True)
            self.cleaned = True
        return self._proc.poll() is not None

    def kill(self):
        if not self.killed:
            try:
                self._proc.kill()
            except:
                pass
            self.cleaned = True
            self.killed = True

    def terminate(self):
        # this shouldn't block, since terminate() is called only
        # after the poller has decided that the worker is dead
        returncode = self._proc.wait()

        # consume all remaining loglines
        # we set the file descriptor to be non-blocking, since
        # the pipe may stay active due to subprocesses launched by
        # the worker, e.g. sidecars, so we can't rely on EOF. We try to
        # read just what's available in the pipe buffer
        for fileobj, buf in self._logs.values():
            fileno = fileobj.fileno()
            fcntl.fcntl(fileno, fcntl.F_SETFL, os.O_NONBLOCK)
            try:
                while self.read_logline(fileno):
                    pass
            except:
                # ignore "resource temporarily unavailable" etc. errors
                # caused due to non-blocking. Draining is done on a
                # best-effort basis.
                pass

        # Return early if the task is cloned since we don't want to
        # perform any log collection.
        if not self.task.is_cloned:
            self.task.save_metadata(
                "runtime",
                {
                    "return_code": returncode,
                    "killed": self.killed,
                    "success": returncode == 0,
                },
            )
            if returncode:
                if not self.killed:
                    if returncode == -11:
                        self.emit_log(
                            b"Task failed with a segmentation fault.",
                            self._stderr,
                            system_msg=True,
                        )
                    else:
                        self.emit_log(b"Task failed.", self._stderr, system_msg=True)
            else:
                num = self.task.results["_foreach_num_splits"]
                if num:
                    self.task.log(
                        "Foreach yields %d child steps." % num,
                        system_msg=True,
                        pid=self._proc.pid,
                    )
                self.task.log(
                    "Task finished successfully.", system_msg=True, pid=self._proc.pid
                )
            self.task.save_logs(
                {
                    "stdout": self._stdout.get_buffer(),
                    "stderr": self._stderr.get_buffer(),
                }
            )

        return returncode

    def __str__(self):
        return "Worker[%d]: %s" % (self._proc.pid, self.task.path)



import inspect
import ast
import re

from itertools import chain


from .util import to_pod


def deindent_docstring(doc):
    if doc:
        # Find the indent to remove from the docstring. We consider the following possibilities:
        # Option 1:
        #  """This is the first line
        #    This is the second line
        #  """
        # Option 2:
        #  """
        # This is the first line
        # This is the second line
        # """
        # Option 3:
        #  """
        #     This is the first line
        #     This is the second line
        #  """
        #
        # In all cases, we can find the indent to remove by doing the following:
        #  - Check the first non-empty line, if it has an indent, use that as the base indent
        #  - If it does not have an indent and there is a second line, check the indent of the
        #    second line and use that
        saw_first_line = False
        matched_indent = None
        for line in doc.splitlines():
            if line:
                matched_indent = re.match("[\t ]+", line)
                if matched_indent is not None or saw_first_line:
                    break
                saw_first_line = True
        if matched_indent:
            return re.sub(r"\n" + matched_indent.group(), "\n", doc).strip()
        else:
            return doc
    else:
        return ""


class DAGNode(object):
    def __init__(
        self, func_ast, decos, wrappers, config_decorators, doc, source_file, lineno
    ):
        self.name = func_ast.name
        self.source_file = source_file
        # lineno is the start line of decorators in source_file
        # func_ast.lineno is lines from decorators start to def of function
        self.func_lineno = lineno + func_ast.lineno - 1
        self.decorators = decos
        self.wrappers = wrappers
        self.config_decorators = config_decorators
        self.doc = deindent_docstring(doc)
        self.parallel_step = any(getattr(deco, "IS_PARALLEL", False) for deco in decos)

        # these attributes are populated by _parse
        self.tail_next_lineno = 0
        self.type = None
        self.out_funcs = []
        self.has_tail_next = False
        self.invalid_tail_next = False
        self.num_args = 0
        self.switch_cases = {}
        self.condition = None
        self.foreach_param = None
        self.num_parallel = 0
        self.parallel_foreach = False
        self._parse(func_ast, lineno)

        # these attributes are populated by _traverse_graph
        self.in_funcs = set()
        self.split_parents = []
        self.split_branches = []
        self.matching_join = None
        # these attributes are populated by _postprocess
        self.is_inside_foreach = False

    def _expr_str(self, expr):
        return "%s.%s" % (expr.value.id, expr.attr)

    def _parse_switch_dict(self, dict_node):
        switch_cases = {}

        if isinstance(dict_node, ast.Dict):
            for key, value in zip(dict_node.keys, dict_node.values):
                case_key = None

                # handle string literals
                if isinstance(key, ast.Str):
                    case_key = key.s
                elif isinstance(key, ast.Constant):
                    case_key = key.value
                elif isinstance(key, ast.Attribute):
                    if isinstance(key.value, ast.Attribute) and isinstance(
                        key.value.value, ast.Name
                    ):
                        # This handles self.config.some_key
                        if key.value.value.id == "self":
                            config_var = key.value.attr
                            config_key = key.attr
                            case_key = f"config:{config_var}.{config_key}"
                        else:
                            return None
                    else:
                        return None

                # handle variables or other dynamic expressions - not allowed
                elif isinstance(key, ast.Name):
                    return None
                else:
                    # can't statically analyze this key
                    return None

                if case_key is None:
                    return None

                # extract the step name from the value
                if isinstance(value, ast.Attribute) and isinstance(
                    value.value, ast.Name
                ):
                    if value.value.id == "self":
                        step_name = value.attr
                        switch_cases[case_key] = step_name
                    else:
                        return None
                else:
                    return None

        return switch_cases if switch_cases else None

    def _parse(self, func_ast, lineno):
        self.num_args = len(func_ast.args.args)
        tail = func_ast.body[-1]

        # end doesn't need a transition
        if self.name == "end":
            # TYPE: end
            self.type = "end"

        # ensure that the tail an expression
        if not isinstance(tail, ast.Expr):
            return

        # determine the type of self.next transition
        try:
            if not self._expr_str(tail.value.func) == "self.next":
                return

            self.has_tail_next = True
            self.invalid_tail_next = True
            self.tail_next_lineno = lineno + tail.lineno - 1

            # Check if first argument is a dictionary (switch case)
            if (
                len(tail.value.args) == 1
                and isinstance(tail.value.args[0], ast.Dict)
                and any(k.arg == "condition" for k in tail.value.keywords)
            ):
                # This is a switch statement
                switch_cases = self._parse_switch_dict(tail.value.args[0])
                condition_name = None

                # Get condition parameter
                for keyword in tail.value.keywords:
                    if keyword.arg == "condition":
                        if isinstance(keyword.value, ast.Str):
                            condition_name = keyword.value.s
                        elif isinstance(keyword.value, ast.Constant) and isinstance(
                            keyword.value.value, str
                        ):
                            condition_name = keyword.value.value
                        break

                if switch_cases and condition_name:
                    self.type = "split-switch"
                    self.condition = condition_name
                    self.switch_cases = switch_cases
                    self.out_funcs = list(switch_cases.values())
                    self.invalid_tail_next = False
                    return

            else:
                self.out_funcs = [e.attr for e in tail.value.args]

            keywords = dict(
                (k.arg, getattr(k.value, "s", None)) for k in tail.value.keywords
            )
            if len(keywords) == 1:
                if "foreach" in keywords:
                    # TYPE: foreach
                    self.type = "foreach"
                    if len(self.out_funcs) == 1:
                        self.foreach_param = keywords["foreach"]
                        self.invalid_tail_next = False
                elif "num_parallel" in keywords:
                    self.type = "foreach"
                    self.parallel_foreach = True
                    if len(self.out_funcs) == 1:
                        self.num_parallel = keywords["num_parallel"]
                        self.invalid_tail_next = False
            elif len(keywords) == 0:
                if len(self.out_funcs) > 1:
                    # TYPE: split
                    self.type = "split"
                    self.invalid_tail_next = False
                elif len(self.out_funcs) == 1:
                    # TYPE: linear
                    if self.name == "start":
                        self.type = "start"
                    elif self.num_args > 1:
                        self.type = "join"
                    else:
                        self.type = "linear"
                    self.invalid_tail_next = False
        except AttributeError:
            return

    def __str__(self):
        return """*[{0.name} {0.type} ({0.source_file} line {0.func_lineno})]*
    in_funcs={in_funcs}
    out_funcs={out_funcs}
    split_parents={parents}
    split_branches={branches}
    matching_join={matching_join}
    is_inside_foreach={is_inside_foreach}
    decorators={decos}
    num_args={0.num_args}
    has_tail_next={0.has_tail_next} (line {0.tail_next_lineno})
    invalid_tail_next={0.invalid_tail_next}
    foreach_param={0.foreach_param}
    condition={0.condition}
    parallel_step={0.parallel_step}
    parallel_foreach={0.parallel_foreach}
    -> {out}""".format(
            self,
            matching_join=self.matching_join and "[%s]" % self.matching_join,
            is_inside_foreach=self.is_inside_foreach,
            out_funcs=", ".join("[%s]" % x for x in self.out_funcs),
            in_funcs=", ".join("[%s]" % x for x in self.in_funcs),
            parents=", ".join("[%s]" % x for x in self.split_parents),
            branches=", ".join("[%s]" % x for x in self.split_branches),
            decos=" | ".join(map(str, self.decorators)),
            out=", ".join("[%s]" % x for x in self.out_funcs),
        )


class FlowGraph(object):
    def __init__(self, flow):
        self.name = flow.__name__
        self.nodes = self._create_nodes(flow)
        self.doc = deindent_docstring(flow.__doc__)
        # nodes sorted in topological order.
        self.sorted_nodes = []
        self._traverse_graph()
        self._postprocess()

    def _create_nodes(self, flow):
        nodes = {}
        for element in dir(flow):
            func = getattr(flow, element)
            if callable(func) and hasattr(func, "is_step"):
                source_file = inspect.getsourcefile(func)
                source_lines, lineno = inspect.getsourcelines(func)
                # This also works for code (strips out leading whitspace based on
                # first line)
                source_code = deindent_docstring("".join(source_lines))
                function_ast = ast.parse(source_code).body[0]
                node = DAGNode(
                    function_ast,
                    func.decorators,
                    func.wrappers,
                    func.config_decorators,
                    func.__doc__,
                    source_file,
                    lineno,
                )
                nodes[element] = node
        return nodes

    def _postprocess(self):
        # any node who has a foreach as any of its split parents
        # has is_inside_foreach=True *unless* all of those `foreach`s
        # are joined by the node
        for node in self.nodes.values():
            foreaches = [
                p for p in node.split_parents if self.nodes[p].type == "foreach"
            ]
            if [f for f in foreaches if self.nodes[f].matching_join != node.name]:
                node.is_inside_foreach = True

    def _traverse_graph(self):
        def traverse(node, seen, split_parents, split_branches):
            add_split_branch = False
            try:
                self.sorted_nodes.remove(node.name)
            except ValueError:
                pass
            self.sorted_nodes.append(node.name)
            if node.type in ("split", "foreach"):
                node.split_parents = split_parents
                node.split_branches = split_branches
                add_split_branch = True
                split_parents = split_parents + [node.name]
            elif node.type == "split-switch":
                node.split_parents = split_parents
                node.split_branches = split_branches
            elif node.type == "join":
                # ignore joins without splits
                if split_parents:
                    self[split_parents[-1]].matching_join = node.name
                    node.split_parents = split_parents
                    node.split_branches = split_branches[:-1]
                    split_parents = split_parents[:-1]
                    split_branches = split_branches[:-1]
            else:
                node.split_parents = split_parents
                node.split_branches = split_branches

            for n in node.out_funcs:
                # graph may contain loops - ignore them
                if n not in seen:
                    # graph may contain unknown transitions - ignore them
                    if n in self:
                        child = self[n]
                        child.in_funcs.add(node.name)
                        traverse(
                            child,
                            seen + [n],
                            split_parents,
                            split_branches + ([n] if add_split_branch else []),
                        )

        if "start" in self:
            traverse(self["start"], [], [], [])

        # fix the order of in_funcs
        for node in self.nodes.values():
            node.in_funcs = sorted(node.in_funcs)

    def __getitem__(self, x):
        return self.nodes[x]

    def __contains__(self, x):
        return x in self.nodes

    def __iter__(self):
        return iter(self.nodes.values())

    def __str__(self):
        return "\n".join(str(self[n]) for n in self.sorted_nodes)

    def output_dot(self):
        def edge_specs():
            for node in self.nodes.values():
                if node.type == "split-switch":
                    # Label edges for switch cases
                    for case_value, step_name in node.switch_cases.items():
                        yield (
                            '{0} -> {1} [label="{2}" color="blue" fontcolor="blue"];'.format(
                                node.name, step_name, case_value
                            )
                        )
                else:
                    for edge in node.out_funcs:
                        yield "%s -> %s;" % (node.name, edge)

        def node_specs():
            for node in self.nodes.values():
                if node.type == "split-switch":
                    # Hexagon shape for switch nodes
                    condition_label = (
                        f"switch: {node.condition}" if node.condition else "switch"
                    )
                    yield (
                        '"{0.name}" '
                        '[ label = <<b>{0.name}</b><br/><font point-size="9">{condition}</font>> '
                        '  fontname = "Helvetica" '
                        '  shape = "hexagon" '
                        '  style = "filled" fillcolor = "lightgreen" ];'
                    ).format(node, condition=condition_label)
                else:
                    nodetype = "join" if node.num_args > 1 else node.type
                    yield '"{0.name}"' '[ label = <<b>{0.name}</b> | <font point-size="10">{type}</font>> ' '  fontname = "Helvetica" ' '  shape = "record" ];'.format(
                        node, type=nodetype
                    )

        return (
            "digraph {0.name} {{\n"
            "{nodes}\n"
            "{edges}\n"
            "}}".format(
                self, nodes="\n".join(node_specs()), edges="\n".join(edge_specs())
            )
        )

    def output_steps(self):
        steps_info = {}
        graph_structure = []

        def node_to_type(node):
            if node.type in ["linear", "start", "end", "join"]:
                return node.type
            elif node.type == "split":
                return "split-static"
            elif node.type == "foreach":
                if node.parallel_foreach:
                    return "split-parallel"
                return "split-foreach"
            elif node.type == "split-switch":
                return "split-switch"
            return "unknown"  # Should never happen

        def node_to_dict(name, node):
            d = {
                "name": name,
                "type": node_to_type(node),
                "line": node.func_lineno,
                "source_file": node.source_file,
                "doc": node.doc,
                "decorators": [
                    {
                        "name": deco.name,
                        "attributes": to_pod(deco.attributes),
                        "statically_defined": deco.statically_defined,
                        "inserted_by": deco.inserted_by,
                    }
                    for deco in node.decorators
                    if not deco.name.startswith("_")
                ]
                + [
                    {
                        "name": deco.decorator_name,
                        "attributes": {"_args": deco._args, **deco._kwargs},
                        "statically_defined": deco.statically_defined,
                        "inserted_by": deco.inserted_by,
                    }
                    for deco in chain(node.wrappers, node.config_decorators)
                ],
                "next": node.out_funcs,
            }
            if d["type"] == "split-foreach":
                d["foreach_artifact"] = node.foreach_param
            elif d["type"] == "split-parallel":
                d["num_parallel"] = node.num_parallel
            elif d["type"] == "split-switch":
                d["condition"] = node.condition
                d["switch_cases"] = node.switch_cases
            if node.matching_join:
                d["matching_join"] = node.matching_join
            return d

        def populate_block(start_name, end_name):
            cur_name = start_name
            resulting_list = []
            while cur_name != end_name:
                cur_node = self.nodes[cur_name]
                node_dict = node_to_dict(cur_name, cur_node)

                steps_info[cur_name] = node_dict
                resulting_list.append(cur_name)

                node_type = node_to_type(cur_node)
                if node_type in ("split-static", "split-foreach"):
                    resulting_list.append(
                        [
                            populate_block(s, cur_node.matching_join)
                            for s in cur_node.out_funcs
                        ]
                    )
                    cur_name = cur_node.matching_join
                elif node_type == "split-switch":
                    all_paths = [
                        populate_block(s, end_name)
                        for s in cur_node.out_funcs
                        if s != cur_name
                    ]
                    resulting_list.append(all_paths)
                    cur_name = end_name
                else:
                    # handles only linear, start, and join steps.
                    if cur_node.out_funcs:
                        cur_name = cur_node.out_funcs[0]
                    else:
                        # handles terminal nodes or when we jump to 'end_name'.
                        break
            return resulting_list

        graph_structure = populate_block("start", "end")

        steps_info["end"] = node_to_dict("end", self.nodes["end"])
        graph_structure.append("end")

        return steps_info, graph_structure
